crop_size = (320, 768)

sintel_global_transform = dict(
    translates=(0.05, 0.05),
    zoom=(1.0, 1.2),
    shear=(0.95, 1.1),
    rotate=(-5., 5.))

sintel_relative_transform = dict(
    translates=(0.00375, 0.00375),
    zoom=(0.985, 1.015),
    shear=(1.0, 1.0),
    rotate=(-1.0, 1.0))

sintel_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', with_occ=True),
    dict(
        type='ColorJitter',
        brightness=0.5,
        contrast=0.5,
        saturation=0.5,
        hue=0.5),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=sintel_global_transform,
        relative_transform=sintel_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='PackFlowInputs')
]

sintel_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputResize', exponent=4),
    dict(type='PackFlowInputs')
]

sintel_clean_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')

sintel_final_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

sintel_clean_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')

sintel_final_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

kitti_global_transform = dict(
    translates=(0.02, 0.02),
    zoom=(0.98, 1.02),
    shear=(1.0, 1.0),
    rotate=(-0.5, 0.5))

kitti_relative_transform = dict(
    translates=(0.0025, 0.0025),
    zoom=(0.99, 1.01),
    shear=(1.0, 1.0),
    rotate=(-0.5, 0.5))

kitti_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(
        type='ColorJitter',
        brightness=0.05,
        contrast=0.2,
        saturation=0.25,
        hue=0.1),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=kitti_global_transform,
        relative_transform=kitti_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='PackFlowInputs')
]

kitti2015_train = dict(
    type='KITTI2015',
    data_root='data/kitti2015',
    pipeline=kitti_train_pipeline,
    test_mode=False)

hd1k_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(type='RandomCrop', crop_size=(436, 1024)),
    dict(
        type='ColorJitter',
        brightness=0.05,
        contrast=0.2,
        saturation=0.25,
        hue=0.1),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=kitti_global_transform,
        relative_transform=kitti_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='PackFlowInputs')
]

hd1k_train = dict(
    type='HD1K',
    data_root='data/hd1k',
    pipeline=hd1k_train_pipeline,
    test_mode=False)

sintel_train = dict(
    type='ConcatDataset', datasets=[sintel_clean_train, sintel_final_train])

train_dataloader = dict(
    batch_size=1,
    num_workers=2,
    sampler=dict(
        type='MixedBatchDistributedSampler',
        sample_ratio=[0.5, 0.25, 0.25],
        shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=dict(
        type='ConcatDataset',
        datasets=[sintel_train, kitti2015_train, hd1k_train]))

val_dataloader = [
    dict(
        batch_size=1,
        num_workers=5,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_clean_test),
    dict(
        batch_size=1,
        num_workers=5,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_final_test)
]

test_dataloader = val_dataloader
val_evaluator = [
    dict(type='EndPointError', prefix='clean'),
    dict(type='EndPointError', prefix='final')
]
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from .base_encoder import BasicConvBlock, BasicEncoder
from .flownet_encoder import CorrEncoder, FlowNetEncoder, FlowNetSDEncoder
from .liteflownet_encoder import NetC
from .pwcnet_encoder import PWCNetEncoder
from .raft_encoder import RAFTEncoder

__all__ = [
    'FlowNetEncoder', 'PWCNetEncoder', 'RAFTEncoder', 'NetC',
    'FlowNetSDEncoder', 'CorrEncoder', 'BasicConvBlock', 'BasicEncoder'
]



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp

import cv2
import mmcv
import numpy as np
from mmengine.utils import scandir

try:
    import imageio
except ImportError:
    imageio = None


def parse_args():
    parser = argparse.ArgumentParser(
        description='Merge images and visualized flow')
    parser.add_argument(
        '--img_dir', type=str, default=None, help='directory of images')
    parser.add_argument(
        '--flow_dir',
        type=str,
        default=None,
        help='directory of visualized flow')
    parser.add_argument(
        '--resize_factor',
        type=float,
        default=0.5,
        help='resize factor for gif')
    parser.add_argument(
        '--out_dir',
        type=str,
        default=None,
        help='directory to save merged results')

    args = parser.parse_args()
    return args


def merge_imgs_flow(img_dir: str, flow_dir: str, out_dir: str) -> None:
    """Load images and visualized flow maps and merge them.

    Args:
        img_dir ([str): The directory of images.
        flow_dir (str): The directory of flow maps.
        out_dir (str): The directory to save the frames
    """
    img_files = list(scandir(img_dir))
    flow_files = list(scandir(flow_dir))
    img_files.sort()
    flow_files.sort()
    # img is longer than flow
    for i in range(len(img_files) - 1):
        img = mmcv.imread(osp.join(img_dir, img_files[i]))
        flow = mmcv.imread(osp.join(flow_dir, flow_files[i]))
        frame = np.concatenate((img, flow), axis=1)

        cv2.imwrite(osp.join(out_dir, flow_files[i]), frame)


def main():
    args = parse_args()
    merge_imgs_flow(args.img_dir, args.flow_dir, args.out_dir)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptMultiConfig
from ..builder import build_components


@MODELS.register_module()
class FlowRefine(BaseModule):
    """Bilateral refinement module for flow in IRR.

    Use a feature-driven local convolution to regularize flow for smoothing
    flow field.

    Args:
        in_channels (int): Number of input channels.
        feat_channels (Sequence[int]): List of numbers of outputs feature
            channels. Default: (128, 128, 96, 64, 64, 32, 32).
        patch_size (int): The size of regularization filter that works for
            feature-driven local convolution. Default: 3.
        warp_cfg (dict): Config for warp Operation. Default: dict(type='Warp').
        conv_cfg (dict , optional): Config for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config for activation layer in ConvModule.
            sDefault: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: Sequence[int] = (128, 128, 96, 64, 64, 32, 32),
                 patch_size: int = 3,
                 warp_cfg: dict = dict(type='Warp', align_corners=True),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(init_cfg)

        self.in_channels = in_channels
        self.feat_channels = feat_channels
        self.patch_size = patch_size
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.warp_op = build_components(warp_cfg)

        layers = []
        for ch in self.feat_channels:
            layers.append(
                ConvModule(
                    in_channels=in_channels,
                    out_channels=ch,
                    kernel_size=3,
                    stride=1,
                    dilation=1,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channels = ch
        self.feat_out_channels = feat_channels[-1]
        layers.append(
            ConvModule(
                in_channels=self.feat_out_channels,
                # patch_size × patch_size regularization filter
                out_channels=patch_size * patch_size,
                kernel_size=3,
                stride=1,
                padding=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg))
        self.layers = nn.Sequential(*layers)

    def forward(
        self,
        img1: Tensor,
        img2: Tensor,
        feat: Tensor,
        flow: Tensor,
    ) -> Tensor:
        """Forward function for IRR-PWC.

        Args:
            img1 (Tensor): The first input image1.
            img2 (Tensor): The second input image2.
            feat (Tensor): The input feature from 1x1 convolution layer in
                Correlation block.
            flow (Tensor): The flow needed to be refined.

        Returns:
            Tensor: The refined optical flow.
        """
        B = img1.shape[0]

        warp_img2 = self.warp_op(img2, flow)
        diff_img = torch.norm((img1 - warp_img2), dim=1, p=2, keepdim=True)
        nomean_flow = flow - flow.view(B, 2, -1).mean(2, True).view(B, 2, 1, 1)
        feat = torch.cat((nomean_flow, diff_img, feat), dim=1)

        # shape (B, patch_size × patch_size, H, W)
        feat = self.layers(feat)
        # a feature-driven CNN distance metric normalized with softmax
        feat = F.softmax(-feat**2, dim=1)

        # flow[:, 0, ...].shape is [B, H, W]
        # flow[:, 0:1, ...].shape is [B, 1, H, W]

        # Extracts sliding local blocks with shape
        # (B, patch_size × patch_size, H, W)
        # from horizontal flow map with shape(B, 1, H, W)
        flow_x_unfold = F.unfold(
            F.pad(
                flow[:, 0:1, ...],
                pad=[self.patch_size // 2] * 4,
                mode='replicate'),
            kernel_size=self.patch_size).view_as((feat))

        flow_y_unfold = F.unfold(
            F.pad(
                flow[:, 1:2, ...],
                pad=[self.patch_size // 2] * 4,
                mode='replicate'),
            kernel_size=self.patch_size).view_as(feat)

        # local convolution
        flow_x = torch.sum(feat * flow_x_unfold, dim=1, keepdim=True)
        flow_y = torch.sum(feat * flow_y_unfold, dim=1, keepdim=True)

        return torch.cat((flow_x, flow_y), dim=1)


@MODELS.register_module()
class OccRefine(FlowRefine):
    """Bilateral refinement module for occlusion in IRR.

    Args:
        in_channels (int): Number of input channels.
        feat_channels (Sequence[int]): List of numbers of outputs feature
            channels. Default: (128, 128, 96, 64, 64, 32, 32).
        patch_size (int): The size of regularization filter that works for
            feature-driven local convolution. Default: 3.
        warp_cfg (dict): Config for warp Operation. Default: dict(type='Warp').
        conv_cfg (dict , optional): Config for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config for activation layer in ConvModule.
            sDefault: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: Sequence[int] = (128, 128, 96, 64, 64, 32, 32),
                 patch_size: int = 3,
                 warp_cfg=dict(type='Warp'),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(
            in_channels=in_channels,
            feat_channels=feat_channels,
            patch_size=patch_size,
            norm_cfg=norm_cfg,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg,
            warp_cfg=warp_cfg,
            init_cfg=init_cfg)

    def forward(
        self,
        feat1: Tensor,
        feat2: Tensor,
        occ: Tensor,
        flow: Tensor,
    ) -> Tensor:
        """Forward function of OccRefine.

        Args:
            feat1 (Tensor): Input feature extracted from the first image.
            feat2 (Tensor): Input feature extracted from the second image.
            occ (Tensor): Current estimated occlusion mask.
            flow (Tensor): Current estimated optical flow.

        Returns:
            Tensor: The occlusion mask after refining.
        """

        warp_feat2 = self.warp_op(feat2, flow)
        diff_feat = feat1 - warp_feat2

        feat = torch.cat((occ, feat1, diff_feat), dim=1)
        feat = self.layers(feat)
        feat = F.softmax(-feat**2, dim=1)

        occ_unfold = F.unfold(
            F.pad(occ, pad=[self.patch_size // 2] * 4, mode='replicate'),
            kernel_size=self.patch_size).view_as((feat))

        occ = torch.sum(feat * occ_unfold, dim=1, keepdim=True)

        return occ


@MODELS.register_module()
class OccShuffleUpsample(BaseModule):
    """Refine module for upsampled occlusion output.

    Args:
        in_channels (int): Number of input channels.
        feat_channels (int): Number of feature channels in residual block.
        infeat_channels (int): Number of input feature channels for conv1x1.
            For reusing this module for different levels, it needs a same input
            channels, and conv1x1 works for modify the input channels to 3.
        out_channels (int): Number of output channels.
        warp_cfg (dict): Config for warp Operation. Default: dict(type='Warp').
        conv_cfg (dict , optional): Config for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config for activation layer in ConvModule.
            sDefault: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: int,
                 infeat_channels: int,
                 out_channels: int,
                 warp_cfg: dict = dict(type='Warp'),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: OptMultiConfig = None) -> None:

        super().__init__(init_cfg=init_cfg)

        self.in_channels = in_channels
        self.feat_channels = feat_channels
        self.out_channels = out_channels
        self.init_conv = ConvModule(
            in_channels=in_channels,
            out_channels=feat_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.res_conv = nn.Sequential(
            ConvModule(
                in_channels=feat_channels,
                out_channels=feat_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg),
            ConvModule(
                in_channels=feat_channels,
                out_channels=feat_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None))
        self.res_end_conv = ConvModule(
            in_channels=feat_channels,
            out_channels=feat_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.out_conv = ConvModule(
            in_channels=feat_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        self.mul_const = 0.1
        self.warp_op = build_components(warp_cfg)
        self.conv_1x1 = ConvModule(
            in_channels=infeat_channels,
            out_channels=3,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

    def forward(
        self,
        occ: Tensor,
        feat1: Tensor,
        feat2: Tensor,
        flow_f: Tensor,
        flow_b: Tensor,
        flow_div: Tensor,
        H_img: Tensor,
        W_img: Tensor,
    ) -> Tensor:
        """Forward function of OccShuffleUpsample.

        Args:
            occ (Tensor): Current estimated occlusion mask.
            feat1 (Tensor): Input feature extracted from the first image.
            feat2 (Tensor): Input feature extracted from the second image.
            flow_f (Tensor): Current estimated optical flow from the first
                image to the second image.
            flow_f (Tensor): Current estimated optical flow from the second
                image to the first image.
            flow_div (Tensor): The divisor to scale optical flow.
            H_img (Tensor): The height of input images.
            W_img (Tensor): The width of input images.

        Returns:
            Tensor: Occlusion mask after refining,
        """

        feat2_warp = self.warp_op(feat2, flow_f)

        h_org, w_org = flow_f.shape[2:]
        u_scale = float(W_img) / float(w_org)
        v_scale = float(H_img) / float(h_org)

        flow_b_ = torch.zeros_like(flow_b)
        flow_b_[:, 0, ...] = flow_b[:, 0, ...] * u_scale / flow_div
        flow_b_[:, 1, ...] = flow_b[:, 1, ...] * v_scale / flow_div

        flow_b_warp = self.warp_op(flow_b_, flow_f)

        if feat1.shape[1] > 3:
            feat1 = self.conv_1x1(feat1)
            feat2_warp = self.conv_1x1(feat2_warp)

        flow_f_ = torch.zeros_like(flow_f)
        flow_f_[:, 0, ...] = flow_f[:, 0, ...] * u_scale / flow_div
        flow_f_[:, 1, ...] = flow_f[:, 1, ...] * v_scale / flow_div

        # feed a occlusion map, a feature map from the encoder,
        # a warped feature map from the other temporal direction,
        # flow, and warped flow.
        feat = torch.cat((occ, feat1, feat2_warp, flow_f_, flow_b_warp), dim=1)
        feat_init = self.init_conv(feat)
        feat_res = feat_init
        feat_res = feat_res + self.res_conv(feat_res) * self.mul_const
        feat_res = feat_res + self.res_conv(feat_res) * self.mul_const
        feat_res = feat_res + self.res_conv(feat_res) * self.mul_const
        feat_init = feat_init + self.res_end_conv(feat_res)

        return self.out_conv(feat_init) + occ



img_norm_cfg = dict(mean=[0., 0., 0.], std=[255., 255., 255.], to_rgb=False)

kitti_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(type='InputResize', exponent=6),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='TestFormatBundle'),
    dict(
        type='Collect',
        keys=['imgs'],
        meta_keys=[
            'flow_gt', 'valid', 'filename1', 'filename2', 'ori_filename1',
            'ori_filename2', 'ori_shape', 'img_shape', 'img_norm_cfg',
            'scale_factor', 'pad_shape'
        ])
]

kitti2015_val_test = dict(
    type='KITTI2015',
    data_root='data/kitti2015',
    pipeline=kitti_test_pipeline,
    test_mode=True)

kitti2012_val_test = dict(
    type='KITTI2012',
    data_root='data/kitti2012',
    pipeline=kitti_test_pipeline,
    test_mode=True)

data = dict(
    test_dataloader=dict(samples_per_gpu=1, workers_per_gpu=2, shuffle=False),
    test=dict(
        type='ConcatDataset',
        datasets=[kitti2012_val_test, kitti2015_val_test],
        separate_eval=True))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch
import torch.nn as nn

from mmflow.registry import MODELS
from .multilevel_flow_loss import multi_level_flow_loss


def charbonnier_loss(pred: torch.Tensor,
                     target: torch.Tensor,
                     q: float = 0.2,
                     eps: float = 0.01) -> torch.Tensor:
    """Generalized Charbonnier loss function between output and ground truth.

    The loss function is
    .. math::
      loss = ((u-u_gt)^2+(v-v_gt)^2+eps)^q

    Generalized Charbonnier loss was used in LiteFlowNet when fine tuning,
    with eps=0.01 q=0.2.

    Args:
        pred (torch.Tensor): output flow map from flow_estimator
            shape(B, 2, H, W).
        target (torch.Tensor): ground truth flow map shape(B, 2, H, W).
        q (float): the exponent in charbonnier loss.
        eps (float): small constant to numerical stability when
            fine-tuning model. Defaults to 0.01.

    Returns:
        Tensor: loss map with the shape (B, H, W).
    """

    assert pred.shape == target.shape, \
        (f'pred shape {pred.shape} does not match target '
         f'shape {target.shape}.')

    diff = torch.add(pred, -target)

    loss_map = (torch.sum(diff * diff, dim=1) + eps)**q  # shape (B, H, W).

    return loss_map


@MODELS.register_module()
class MultiLevelCharbonnierLoss(nn.Module):
    """Multi-level Generalized Charbonnier loss.

    Args:

        q (float): the exponent in charbonnier loss.
        eps (float): small constant to numerical stability when
            fine-tuning model. Defaults to 0.01.
        weights (dict): manual rescaling weights given to the loss of flow map
            at each level, and the keys in weights must correspond to predicted
            dict. Defaults to dict(
            level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005).
        flow_div (float): the divisor used to scale down ground truth.
            Defaults to 20.
        max_flow (float): maximum value of optical flow, if some pixel's flow
            of target is larger than it, this pixel is not valid. Default to
            float('inf').
        resize_flow (str): mode for reszing flow: 'downsample' and 'upsample',
            as multi-level predicted outputs don't match the ground truth.
            If set to 'downsample', it will downsample the ground truth, and
            if set to 'upsample' it will upsample the predicted flow, and
            'upsample' is used for sparse flow map as no generic interpolation
            mode can resize a ground truth of sparse flow correctly.
            Default to 'downsample'.
        scale_as_level (bool): Whether flow for each level is at its native
            spatial resolution. If `'scale_as_level'` is True, the ground
            truth is scaled at different levels, if it is False, the ground
            truth will not be scaled. Default to False.
        reduction (str): the reduction to apply to the output:'none', 'mean',
            'sum'. 'none': no reduction will be applied and will return a
            full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
            the epe map will be summed but averaged by batch_size.
            Default: 'sum'.
    """

    def __init__(self,
                 q: float = 0.2,
                 eps: float = 0.01,
                 flow_div: float = 20.,
                 weights: Dict[str, float] = dict(
                     level6=0.32,
                     level5=0.08,
                     level4=0.02,
                     level3=0.01,
                     level2=0.005),
                 max_flow: float = float('inf'),
                 resize_flow: str = 'downsample',
                 scale_as_level: bool = False,
                 reduction: str = 'sum') -> None:
        super().__init__()

        assert isinstance(q, float) and q > 0.
        self.q = q

        assert isinstance(eps, float) and eps > 0.
        self.eps = eps

        assert flow_div > 0
        self.flow_div = flow_div

        assert isinstance(weights, dict)
        self.weights = weights

        assert max_flow > 0.
        self.max_flow = max_flow

        assert resize_flow in ('downsample', 'upsample')
        self.resize_flow = resize_flow

        assert isinstance(scale_as_level, bool)
        self.scale_as_level = scale_as_level

        assert reduction in ('mean', 'sum')
        self.reduction = reduction

    def forward(self,
                pred: Dict[str, Union[torch.Tensor, List[torch.Tensor]]],
                target: torch.Tensor,
                valid: Optional[torch.Tensor] = None) -> torch.Tensor:
        """Forwar function for MultiLevelCharbonnierLoss.

        Args:
            preds_dict (dict): Multi-level output of predicted optical flow,
                and the contain in dict is a Tensor or list of Tensor with
                shape (B, 1, H_l, W_l), where l indicates the level.
            target (Tensor): Ground truth of optical flow with shape
                (B, 2, H, W).
            valid (Tensor, optional): Valid mask for optical flow.
                Defaults to None.

        Returns:
            Tensor: value of pixel-wise generalized Charbonnier loss.
        """

        return multi_level_flow_loss(
            charbonnier_loss,
            pred,
            target,
            weights=self.weights,
            valid=valid,
            flow_div=self.flow_div,
            max_flow=self.max_flow,
            resize_flow=self.resize_flow,
            scale_as_level=self.scale_as_level,
            reduction=self.reduction,
            q=self.q,
            eps=self.eps,
        )

    def __repr__(self) -> str:

        repr_str = self.__class__.__name__
        repr_str += (f'(resize_flow={self.resize_flow}, '
                     f'scale_as_level={self.scale_as_level}, '
                     f'flow_div={self.flow_div}, '
                     f'weights={self.weights}, '
                     f'q={self.q}, '
                     f'eps={self.eps}, '
                     f'reduction=\'{self.reduction}\')')

        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import os.path as osp
from glob import glob
from typing import Callable, List, Optional, Sequence, Union

from mmengine.dataset import BaseDataset as MME_BaseDataset

from mmflow.registry import DATASETS
from .utils import get_data_filename, load_ann_info, load_img_info


@DATASETS.register_module()
class FlyingThings3D(MME_BaseDataset):
    """FlyingThings3D subset dataset.

    Args:

        scene (str, optional): Scene in Flyingthings3D dataset, which has 3
            options None, 'left' and 'right', and None means ``data_list``
            will not be filter by scene filed. Defaults to 'left'.
            This default value is for RAFT, as FlyingThings3D is so large
            and not often used, and only RAFT use the 'left' data in it.
        pass_style (str): Pass style for FlyingThing3D dataset, and it has 2
            options 'clean', 'final'. Default: 'clean'.
        double (bool): Whether double dataset by change the images pairs from
            (img1, img2) to (img2, img1). Defaults to False.
        ann_file (str): Annotation file path. Defaults to ''.
        metainfo (dict, optional): Meta information for dataset, such as class
            information. Defaults to None.
        data_root (str, optional): The root directory for ``data_prefix`` and
            ``ann_file``. Defaults to None.
        data_prefix (dict, optional): Prefix for training data. Defaults to
            dict(img=None, ann=None).
        filter_cfg (dict, optional): Config for filter data. Defaults to None.
        indices (int or Sequence[int], optional): Support using first few
            data in annotation file to facilitate training/testing on a smaller
            dataset. Defaults to None which means using all ``data_infos``.
        serialize_data (bool, optional): Whether to hold memory using
            serialized objects, when enabled, data loader workers can use
            shared RAM from master process instead of making a copy. Defaults
            to True.
        pipeline (list, optional): Processing pipeline. Defaults to [].
        test_mode (bool, optional): ``test_mode=True`` means in test phase.
            Defaults to False.
        lazy_init (bool, optional): Whether to load annotation during
            instantiation. In some cases, such as visualization, only the meta
            information of the dataset is needed, which is not necessary to
            load annotation file. ``Basedataset`` can skip load annotations to
            save time by set ``lazy_init=False``. Defaults to False.
        max_refetch (int, optional): If ``Basedataset.prepare_data`` get a
            None img. The maximum extra number of cycles to get a valid
            image. Defaults to 1000.
    """
    METAINFO = dict()

    def __init__(self,
                 scene: Optional[str] = None,
                 pass_style: str = 'clean',
                 double: bool = False,
                 ann_file: str = '',
                 metainfo: Optional[dict] = None,
                 data_root: Optional[str] = None,
                 data_prefix: dict = dict(),
                 filter_cfg: Optional[dict] = None,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 serialize_data: bool = True,
                 pipeline: List[Union[dict, Callable]] = [],
                 test_mode: bool = False,
                 lazy_init: bool = False,
                 max_refetch: int = 1000) -> None:

        if scene is not None:
            assert scene in (
                'left', 'right'
            ), f'`scene` expected to be \'left\' or \'right\' but got {scene}'
        self.scene = scene

        assert pass_style in (
            'clean',
            'final'), '`pass_style` expected to be \'clean\' or \'final\' '
        f'but got {pass_style}'
        self.pass_style = pass_style
        self.double = double

        super().__init__(ann_file, metainfo, data_root, data_prefix,
                         filter_cfg, indices, serialize_data, pipeline,
                         test_mode, lazy_init, max_refetch)

    def load_data_list(self) -> List[dict]:
        """Load ``data_list``

        ``data_list`` can be load from an annotation file named as
        ``self.ann_file`` or by parsing dataset path.

        Returns:
            list[dict]: A list of annotation.
        """
        if self.ann_file.endswith('json'):
            # load data_list with annotation file
            return super().load_data_list()
        else:
            # load data_list by path parsing
            self.load_data_info()
            return self.data_list

    def filter_data(self) -> List[dict]:
        """Filter ``data_list`` according to ``scene``, ``pass_style`` and
        ``double``

        Returns:
            list[int]: Filtered results.
        """
        if self.scene is None:
            new_data_list = [
                data_info for data_info in self.data_list
                if data_info['pass_style'] == self.pass_style
            ]

        else:

            new_data_list = [
                data_info for data_info in self.data_list
                if data_info['scene'] == self.scene
                and data_info['pass_style'] == self.pass_style
            ]
        if self.double:
            changed_new_data_list = []
            for data_info in new_data_list:
                changed_data_info = copy.deepcopy(data_info)
                changed_data_info['img1_path'], changed_data_info[
                    'img2_path'] = data_info['img2_path'], data_info[
                        'img1_path']
                changed_data_info['flow_fw_path'], changed_data_info[
                    'flow_bw_path'] = data_info['flow_bw_path'], data_info[
                        'flow_fw_path']
                changed_new_data_list.append(changed_data_info)
            return new_data_list + changed_new_data_list

        else:
            return new_data_list

    def load_data_info(self) -> None:
        """Load data information, including file path of image1, image2 and
        optical flow."""
        self._get_data_dir()

        img1_filenames = []
        img2_filenames = []
        flow_fw_filenames = []
        flow_bw_filenames = []

        for idir, fw_dir, bw_dir in zip(self.img1_dir, self.flow_fw_dir,
                                        self.flow_bw_dir):

            img_filenames_ = get_data_filename(idir, self.img1_suffix)
            img_filenames_.sort()
            flow_fw_filenames_ = get_data_filename(fw_dir, self.flow_suffix)
            flow_fw_filenames_.sort()
            flow_bw_filenames_ = get_data_filename(bw_dir, self.flow_suffix)
            flow_bw_filenames_.sort()

            img1_filenames += img_filenames_[:-1]
            img2_filenames += img_filenames_[1:]
            flow_fw_filenames += flow_fw_filenames_[:-1]
            flow_bw_filenames += flow_bw_filenames_[1:]

        load_img_info(self.data_list, img1_filenames, img2_filenames)

        load_ann_info(self.data_list, flow_fw_filenames, 'flow_fw_path')
        load_ann_info(self.data_list, flow_bw_filenames, 'flow_bw_path')

        for i in range(len(self.data_list)):
            self.data_list[i]['pass_style'] = self.pass_style
            self.data_list[i]['scene'] = 'left' if 'left' in self.data_list[i][
                'img1_path'] else 'right'

    def _get_data_dir(self) -> None:
        """Get the paths for images and optical flow."""

        self.flow_fw_dir = 'into_future'
        self.flow_bw_dir = 'into_past'
        self.flow_suffix = '.pfm'
        self.occ_suffix = '.png'
        self.img1_suffix = '.png'
        self.img2_suffix = '.png'

        self.subset_dir = 'TEST' if self.test_mode else 'TRAIN'

        imgs_dirs = []
        flow_fw_dirs = []
        flow_bw_dirs = []

        if self.scene is None:
            scene = ['left', 'right']
        else:
            scene = [self.scene]

        pass_dir = 'frames_' + self.pass_style + 'pass'
        for i_scene in scene:
            imgs_dirs_ = glob(
                osp.join(self.data_root, pass_dir, self.subset_dir + '/*/*'))
            imgs_dirs_ = [osp.join(f, i_scene) for f in imgs_dirs_]
            imgs_dirs += imgs_dirs_

            flow_fw_dirs_ = glob(
                osp.join(self.data_root, 'optical_flow',
                         self.subset_dir + '/*/*/' + self.flow_fw_dir))

            flow_fw_dirs_ = [osp.join(f, i_scene) for f in flow_fw_dirs_]
            flow_fw_dirs += flow_fw_dirs_

            flow_bw_dirs_ = glob(
                osp.join(self.data_root, 'optical_flow',
                         self.subset_dir + '/*/*/' + self.flow_bw_dir))
            flow_bw_dirs_ = [osp.join(f, i_scene) for f in flow_bw_dirs_]
            flow_bw_dirs += flow_bw_dirs_

        self.img1_dir = imgs_dirs
        self.img2_dir = imgs_dirs
        self.flow_fw_dir = flow_fw_dirs
        self.flow_bw_dir = flow_bw_dirs



_base_ = [
    '../_base_/models/irrpwc.py',
    '../_base_/datasets/flyingthings3d_subset_bi_with_occ_384x768.py',
    '../_base_/schedules/schedule_s_fine_half.py',
    '../_base_/default_runtime.py'
]

custom_hooks = [dict(type='EMAHook')]

# Train on FlyingChairsOcc and finetune on FlyingThings3D_subset
load_from = 'https://download.openmmlab.com/mmflow/irr/irrpwc_8x1_sshort_flyingchairsocc_384x448.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence, Union

import torch
from mmcv.cnn import ConvModule

from mmflow.registry import MODELS
from ..utils import CorrBlock
from .base_encoder import BasicEncoder


@MODELS.register_module()
class FlowNetEncoder(BasicEncoder):
    """The feature extraction sub-module of FlowNetS and FlowNetC.

    Args:
        in_channels (int): Number of input channels.
        pyramid_levels (Sequence[str]): The list of feature pyramid that are
            the keys for output dict.
        num_convs (Sequence[int]): Numbers of conv layers for each
            pyramid level. Default: (1, 1, 2, 2, 2, 2).
        out_channels (Sequence[int]): List of numbers of output
            channels of each pyramid level.
            Default: (64, 128, 256, 512, 512, 1024).
        kernel_size (Sequence): List of kernel size of each feature level.
            Default: (7, 5, (5, 3), 3, 3, 3).
        strides (Sequence[int]): List of strides of each pyramid level.
            Default: (2, 2, 2, 2, 2, 2).
        dilations (Sequence[int]): List of dilation of each pyramid level.
            Default: (1, 1, 1, 1, 1, 1).
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict): Config dict for each activation layer in ConvModule.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 pyramid_levels: Sequence[str],
                 num_convs: Sequence[int] = (1, 1, 2, 2, 2, 2),
                 out_channels: Sequence[int] = (64, 128, 256, 512, 512, 1024),
                 kernel_size: Sequence = (7, 5, (5, 3), 3, 3, 3),
                 strides: Sequence[int] = (2, 2, 2, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 1, 1, 1, 1),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[list, dict]] = None) -> None:

        super().__init__(
            in_channels=in_channels,
            pyramid_levels=pyramid_levels,
            num_convs=num_convs,
            out_channels=out_channels,
            kernel_size=kernel_size,
            strides=strides,
            dilations=dilations,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)


@MODELS.register_module()
class CorrEncoder(BasicEncoder):
    """The Correlation feature extraction sub-module of FlowNetC..

    Args:
        in_channels (int): Number of input channels.
        pyramid_levels (Sequence[str]): Number of pyramid levels.
        kernel_sizes (Sequence[int]): List of numbers of kernel size of each
            block. Default: (3, 3, 3, 3).
        num_convs (Sequence[int]): List of number of convolution layers.
            Default: (1, 2, 2, 2).
        out_channels (Sequence[int]): List of numbers of output channels of
            each ConvModule. Default: (256, 512, 512, 1024).
        redir_in_channels (int): Number of input channels of
            redirection ConvModule. Default: 256.
        redir_channels (int): Number of output channels of redirection
            ConvModule. Default: 32.
        strides (Sequence[int]): List of numbers of strides of each block.
            Default: (1, 2, 2, 2).
        dilations (Sequence[int]): List of numbers of dilations of each block.
            Default: (1, 1, 1, 1).
        corr_cfg (dict): Config dict for correlation layer.
            Default: dict(type='Correlation', kernel_size=1, max_displacement
            =10, stride=1, padding=0, dilation_patch=2)
        scaled (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not. Default: False.
        act_cfg (dict): Config for each activation layer in ConvModule.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        conv_cfg (dict, optional): Config for convolution layers.
            Default: None.
        norm_cfg (dict, optional): Config for each normalization layer.
            Default: None.
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 pyramid_levels: Sequence[str],
                 kernel_size: Sequence[int] = (3, 3, 3, 3),
                 num_convs: Sequence[int] = (1, 2, 2, 2),
                 out_channels: Sequence[int] = (256, 512, 512, 1024),
                 redir_in_channels: int = 256,
                 redir_channels: int = 32,
                 strides: Sequence[int] = (1, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 1, 1),
                 corr_cfg: dict = dict(
                     type='Correlation',
                     kernel_size=1,
                     max_displacement=10,
                     stride=1,
                     padding=0,
                     dilation_patch=2),
                 scaled: bool = False,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 init_cfg: Optional[Union[list, dict]] = None) -> None:

        super().__init__(
            in_channels=in_channels,
            pyramid_levels=pyramid_levels,
            num_convs=num_convs,
            out_channels=out_channels,
            kernel_size=kernel_size,
            strides=strides,
            dilations=dilations,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)

        self.corr = CorrBlock(corr_cfg, act_cfg, scaled=scaled)

        self.conv_redir = ConvModule(
            in_channels=redir_in_channels,
            out_channels=redir_channels,
            kernel_size=1,
            act_cfg=act_cfg)

    def forward(self, f1: torch.Tensor,
                f2: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward function for CorrEncoder.

        Args:
            f1 (Tensor): The feature from the first input image.
            f2 (Tensor): The feature from the second input image.

        Returns:
            Dict[str, Tensor]: The feature pyramid for correlation.
        """

        corr_feat = self.corr(f1, f2)
        redir_feat = self.conv_redir(f1)

        x = torch.cat((redir_feat, corr_feat), dim=1)

        outs = dict()
        for i, convs_layer in enumerate(self.layers):
            x = convs_layer(x)
            # After correlation, the feature level starts at level3
            if 'level' + str(i + 3) in self.pyramid_levels:
                outs['level' + str(i + 3)] = x

        return outs


@MODELS.register_module()
class FlowNetSDEncoder(BasicEncoder):
    """The feature extraction sub-module of FlowNetSD.

    Args:
        in_channels (int): Number of input channels.
        plugin_channels (int): The output channels of plugin convolution layer.
        pyramid_levels (Sequence[str]): The list of feature pyramid that are
            the keys for output dict.
        num_convs (Sequence[int]): Numbers of conv layers for each
            pyramid level. Default: (2, 2, 2, 2, 2, 2).
        out_channels (Sequence): List of numbers of output
            channels of each pyramid level.
            Default: ((64, 128), 128, 256, 512, 512, 1024).
        kernel_size (int): Kernel size of each feature level. Default: 3.
        strides (Sequence[int]): List of strides of each pyramid level.
            Default: (2, 2, 2, 2, 2, 2).
        dilations (Sequence[int]): List of dilation of each pyramid level.
            Default: (1, 1, 1, 1, 1, 1).
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict): Config dict for each activation layer in ConvModule.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 plugin_channels: int,
                 pyramid_levels: Sequence[str],
                 num_convs: Sequence[int] = (2, 2, 2, 2, 2, 2),
                 out_channels: Sequence = ((64, 128), 128, 256, 512, 512,
                                           1024),
                 kernel_size: int = 3,
                 strides: Sequence[int] = (2, 2, 2, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 1, 1, 1, 1),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[list, dict]] = None) -> None:
        super().__init__(
            plugin_channels,
            pyramid_levels,
            num_convs=num_convs,
            out_channels=out_channels,
            strides=strides,
            dilations=dilations,
            kernel_size=kernel_size,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)

        self.plugin_layer = ConvModule(
            in_channels=in_channels,
            out_channels=plugin_channels,
            act_cfg=act_cfg,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
        )

    def forward(self, imgs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Forward function for FlowNetSDEncoder.

        Args:
            imgs (Tensor): The concatenate images.

        Returns:
            Dict[str, Tensor]: The feature pyramid extracted from images.
        """
        outs = dict()
        x = self.plugin_layer(imgs)
        for i, convs_layer in enumerate(self.layers):
            x = convs_layer(x)
            if 'level' + str(i + 1) in self.pyramid_levels:
                outs['level' + str(i + 1)] = x

        return outs



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from collections import defaultdict
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import torch
from mmcv.ops import Correlation
from mmengine.config import Config
from mmengine.runner import load_checkpoint

from mmflow.datasets.transforms import Compose
from mmflow.models import build_flow_estimator
from mmflow.structures import FlowDataSample


def init_model(config: Union[str, Config],
               checkpoint: Optional[str] = None,
               device: str = 'cuda:0',
               cfg_options: Optional[dict] = None) -> torch.nn.Module:
    """Initialize a flow estimator from config file.

    Args:
        config (str or :obj:`mmengine.Config`): Config file path or the config
            object.
        checkpoint (str, optional): Checkpoint path. If left as None, the model
            will not load any weights. Default to: None.
        device (str): Represent the device. Default to: 'cuda:0'.
        cfg_options (dict, optional): Options to override some settings in the
            used config. Default to: None.
    Returns:
        nn.Module: The constructed flow estimator.
    """

    if isinstance(config, (str, Path)):
        config = Config.fromfile(config)
    elif not isinstance(config, Config):
        raise TypeError('config must be a filename or Config object, '
                        f'but got {type(config)}')
    if cfg_options is not None:
        config.merge_from_dict(cfg_options)

    config.model.train_cfg = None
    model = build_flow_estimator(config.model)
    if checkpoint is not None:
        checkpoint = load_checkpoint(model, checkpoint, map_location='cpu')

    model.cfg = config  # save the config in the model for convenience
    model.to(device)
    model.eval()

    return model


def inference_model(model: torch.nn.Module, img1s: Union[str, np.ndarray],
                    img2s: Union[str, np.ndarray]) -> List[FlowDataSample]:
    """Inference images pairs with the flow estimator.

    Args:
        model (nn.Module): The loaded flow estimator.
        img1s (str/ndarray or list[str/ndarray] or tuple[str/ndarray]):
           Either image files or loaded images.
        img2s (str/ndarray or list[str/ndarray] or tuple[str/ndarray]):
           Either image files or loaded images.
    Returns:
        If img-pairs is a list or tuple, the same length list type results
        will be returned, otherwise return the flow map from image1 to image2
        directly.
        List[FlowDataSample]: the predicted flows from img1s to img2s.
            For a single img-pair, its prediction is in the data field
            of ``pred_flow_fw`` in FlowDataSample.
    """
    if not isinstance(img1s, (list, tuple)):
        img1s = [img1s]
        img2s = [img2s]

    cfg = model.cfg
    if isinstance(cfg.test_dataloader, list):
        cfg = copy.deepcopy(cfg.test_dataloader[0].dataset)
    else:
        cfg = copy.deepcopy(cfg.test_dataloader.dataset)

    if isinstance(img1s[0], np.ndarray):
        cfg.pipeline[0].type = 'LoadImageFromWebcam'

    # as load annotation is for online evaluation
    # there is no need to load annotation.
    if dict(type='LoadAnnotations') in cfg.pipeline:
        cfg.pipeline.remove(dict(type='LoadAnnotations'))

    test_pipeline = Compose(cfg.pipeline)
    datas = defaultdict(list)
    for img1, img2 in zip(img1s, img2s):
        # prepare data
        if isinstance(img1, np.ndarray) and isinstance(img2, np.ndarray):
            # directly add img
            data = dict(img1=img1, img2=img2)
        else:
            # add information into dict
            data = dict(img1_path=img1, img2_path=img2)
        # build the data pipeline
        data = test_pipeline(data)
        datas['inputs'].append(data['inputs'])
        datas['data_samples'].append(data['data_samples'])

    datas = model.data_preprocessor(datas, False)
    inputs, data_samples = datas['inputs'], datas['data_samples']

    for m in model.modules():
        assert not isinstance(
            m, Correlation
        ), 'CPU inference with Correlation is not supported currently.'

    # forward the model
    with torch.no_grad():
        results = model.predict(inputs, data_samples)

    return results



# training schedule for maskflownet_ft schedule
train_cfg = dict(by_epoch=False, max_iters=1000000, val_interval=50000)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='Adam', lr=5e-5, weight_decay=0.0004, betas=(0.9, 0.999)))

# learning policy
param_scheduler = dict(
    type='MultiStageLR',
    by_epoch=False,
    gammas=[0.5, 0.5, 0.5, 0.5, 0.5],
    milestone_params=[5e-5, 3e-5, 2e-5, 1e-5, 5e-6],
    milestone_iters=[0, 200000, 400000, 600000, 800000],
    steps=[[100000, 150000], [300000, 350000], [500000, 550000],
           [700000, 750000], [850000, 875000, 900000, 950000, 975000]])

default_hooks = dict(
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=50, log_metric_by_epoch=False),
    param_scheduler=dict(type='ParamSchedulerHook'),
    checkpoint=dict(type='CheckpointHook', interval=50000, by_epoch=False),
    sampler_seed=dict(type='DistSamplerSeedHook'),
    visualization=dict(type='FlowVisualizationHook'))



crop_size = (368, 768)

# Sintel config
sintel_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(
        type='ColorJitter',
        asymmetric_prob=0.2,
        brightness=0.4,
        contrast=0.4,
        saturation=0.4,
        hue=0.5 / 3.14),
    dict(type='Erase', prob=0.5, bounds=[50, 100], max_num=3),
    dict(
        type='SpacialTransform',
        spacial_prob=0.8,
        stretch_prob=0.8,
        crop_size=crop_size,
        min_scale=-0.2,
        max_scale=0.6,
        max_stretch=0.2),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.1, direction='vertical'),
    dict(type='Validation', max_flow=1000.),
    dict(type='PackFlowInputs')
]
sintel_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputPad', exponent=3),
    dict(type='PackFlowInputs')
]

sintel_clean_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')
sintel_final_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

sintel_clean_train_x100 = dict(
    type='RepeatDataset', times=100, dataset=sintel_clean_train)
sintel_final_train_x100 = dict(
    type='RepeatDataset', times=100, dataset=sintel_final_train)

sintel_clean_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')
sintel_final_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

# Flyingthings3d config
flyingthing3d_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(
        type='ColorJitter',
        asymmetric_prob=0.2,
        brightness=0.4,
        contrast=0.4,
        saturation=0.4,
        hue=0.5 / 3.14),
    dict(type='Erase', prob=0.5, bounds=[50, 100], max_num=3),
    dict(
        type='SpacialTransform',
        spacial_prob=0.8,
        stretch_prob=0.8,
        crop_size=crop_size,
        min_scale=-0.4,
        max_scale=0.8,
        max_stretch=0.2),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.1, direction='vertical'),
    dict(type='Validation', max_flow=1000.),
    dict(type='PackFlowInputs')
]
flyingthings3d_clean_train = dict(
    type='FlyingThings3D',
    data_root='data/flyingthings3d',
    pipeline=flyingthing3d_train_pipeline,
    test_mode=False,
    pass_style='clean',
    scene='left',
    double=True)

train_dataloader = dict(
    batch_size=2,
    num_workers=5,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=dict(
        type='ConcatDataset',
        datasets=[
            sintel_clean_train_x100, sintel_final_train_x100,
            flyingthings3d_clean_train
        ]))

val_dataloader = [
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_clean_test),
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_final_test)
]
test_dataloader = val_dataloader
val_evaluator = [
    dict(type='EndPointError', prefix='clean'),
    dict(type='EndPointError', prefix='final')
]
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from .end_point_error import EndPointError
from .optical_flow_outliers import FlowOutliers
from .utils import end_point_error_map

__all__ = ['EndPointError', 'FlowOutliers', 'end_point_error_map']



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp

from mmengine.config import Config, DictAction
from mmengine.utils import ProgressBar

from mmflow.datasets.builder import build_dataset
from mmflow.registry import VISUALIZERS
from mmflow.utils import register_all_modules


def parse_args():
    parser = argparse.ArgumentParser(description='Browse a dataset')
    parser.add_argument('config', help='train config file path')
    parser.add_argument('--not-show', default=False, action='store_true')
    parser.add_argument(
        '--show-interval',
        type=float,
        default=2,
        help='the interval of show (s)')
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    cfg = Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    # register all modules in mmseg into the registries
    register_all_modules()

    dataset = build_dataset(cfg.train_dataloader.dataset)

    visualizer = VISUALIZERS.build(cfg.visualizer)
    visualizer.dataset_meta = dataset.METAINFO

    progress_bar = ProgressBar(len(dataset))
    for item in dataset:
        data_sample = item['data_sample']
        img1_path = item['data_sample'].metainfo['img1_path']
        img2_path = item['data_sample'].metainfo['img2_path']

        window_name = f'{osp.basename(img1_path)}_{osp.basename(img2_path)}'

        visualizer.add_datasample(
            window_name,
            None,
            data_sample,
            show=not args.not_show,
            wait_time=args.show_interval)

        progress_bar.update()


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import TensorDict
from .pwcnet import PWCNet


@MODELS.register_module()
class LiteFlowNet(PWCNet):
    """LiteFlowNet model."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract_feat(
            self,
            imgs: Tensor) -> Tuple[Tensor, Tensor, TensorDict, TensorDict]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[Tensor, Tensor, TensorDict, TensorDict]: The
                first input image, the second input image, the feature pyramid
                of the first input image and the feature pyramid of second
                input image.
        """

        in_channels = self.encoder.in_channels

        # take from github.com:sniklaus/pytorch-liteflownet.git
        imgs_mean = [0.411618, 0.434631, 0.454253]

        for ich in range(in_channels):
            imgs[:, ich, :, :] = imgs[:, ich, :, :] - imgs_mean[ich]
            imgs[:, ich + in_channels, :, :] = (
                imgs[:, ich + in_channels, :, :] - imgs_mean[ich])

        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]

        return img1, img2, self.encoder(img1), self.encoder(img2)



# Copyright (c) OpenMMLab. All rights reserved.
from .advanced_transform import RandomAffine
from .compose import Compose
from .formatting import PackFlowInputs
from .loading import LoadAnnotations
from .transforms import (ColorJitter, Erase, GaussianNoise, InputPad,
                         InputResize, Normalize, PhotoMetricDistortion,
                         RandomCrop, RandomFlip, Rerange, SpacialTransform,
                         Validation)

__all__ = [
    'Compose', 'LoadAnnotations', 'SpacialTransform', 'Validation', 'Erase',
    'InputResize', 'InputPad', 'RandomFlip', 'Normalize', 'Rerange',
    'RandomCrop', 'ColorJitter', 'PhotoMetricDistortion', 'GaussianNoise',
    'RandomAffine', 'PackFlowInputs'
]



_base_ = [
    '../_base_/models/flownet2/flownet2css-sd.py',
    '../_base_/datasets/flyingthings3d_subset_chairssdhom_384x448.py',
    '../_base_/schedules/schedule_s_fine.py', '../_base_/default_runtime.py'
]

# Initialize FlowNet2CSS and finetune on FlingThings3D_subset and ChairsSDHom
load_from = 'https://download.openmmlab.com/mmflow/flownet2/flownet2css-sd_8x1_sfine_flyingthings3d_subset_chairssdhom_384x448.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.runner import autocast
from torch import Tensor

from mmflow.registry import MODELS


def coords_grid(batch: int, xx: Tensor, yy: Tensor) -> Tensor:
    """Coordinate grid.

    Args:
        batch (int): The batch size of feature.
        xx (Tensor): 1-D tensor of size W with values from the interval
            [0, W-1].
        yy (Tensor): 1-D tensor of size H with values from the interval
            [0, H-1].

    Returns:
        Tensor: Tensor of shape (batch, 2, H, W) with values of items'
            coordinate.
    """
    coords = torch.meshgrid(yy, xx)
    coords = torch.stack(coords[::-1], dim=0).float()

    return coords[None].repeat(batch, 1, 1, 1)  # shape(batch, 2, H, W)


def bilinear_sample(feat: Tensor,
                    grid: Tensor,
                    mode: str = 'bilinear',
                    padding_mode: str = 'zeros',
                    align_corners: bool = False,
                    scale: bool = True) -> Tensor:
    """Computes the output using input feature values and pixel locations from
    grid.

    Args:
        feat (Tensor): The input feature.
        grid (Tensor): The coordinate grid or the scaled coordinate that has
            values in the range of [-1, 1].
        mode (str): Interpolation mode to calculate output values.
            Defaults to 'bilinear'.
        padding_mode (str): Padding mode for outside grid values.
            Defaults to 'zeros'.
        align_corners (bool): If set to True, the extrema (-1 and 1) are
            considered as referring to the center points of the input’s corner
            pixels. If set to False, they are instead considered as referring
            to the corner points of the input’s corner pixels, making the
            sampling more resolution agnostic. Default to False.
        scale (bool): Whether scale the grid values in the range of [-1, 1].
            Defaults to True.

    Returns:
        Tensor: The output tensor using input feature values and pixel
            locations from grid
    """
    H, W = feat.shape[-2:]
    if grid.shape[-1] != 2:
        grid = grid.permute(0, 2, 3, 1)
    if scale:
        grid[:, :, :, 0] = grid[:, :, :, 0] * 2. / max(W - 1, 1) - 1.
        grid[:, :, :, 1] = grid[:, :, :, 1] * 2. / max(H - 1, 1) - 1.

    with autocast(enabled=False):
        return F.grid_sample(feat.float(), grid, mode, padding_mode,
                             align_corners)


@MODELS.register_module()
class CorrLookup(nn.Module):
    """Correlation lookup operator.

    This operator is used in `RAFT<https://arxiv.org/pdf/2003.12039.pdf>`_

    Args:
        radius (int): the radius of the local neighborhood of the pixels.
            Default to 4.
        mode (str): interpolation mode to calculate output values 'bilinear'
            | 'nearest' | 'bicubic'. Default: 'bilinear' Note: mode='bicubic'
            supports only 4-D input.
        padding_mode (str): padding mode for outside grid values 'zeros' |
            'border' | 'reflection'. Default: 'zeros'
        align_corners (bool): If set to True, the extrema (-1 and 1) are
            considered as referring to the center points of the input’s corner
            pixels. If set to False, they are instead considered as referring
            to the corner points of the input’s corner pixels, making the
            sampling more resolution agnostic. Default to True.
    """

    def __init__(self,
                 radius: int = 4,
                 mode: str = 'bilinear',
                 padding_mode: str = 'zeros',
                 align_corners: bool = True) -> None:
        super().__init__()
        self.r = radius
        self.mode = mode
        self.padding_mode = padding_mode
        self.align_corners = align_corners

    def forward(self, corr_pyramid: Sequence[Tensor], flow: Tensor) -> Tensor:
        """Forward function of Correlation lookup.

        Args:
            corr_pyramid (Sequence[Tensor]): Correlation pyramid.
            flow (Tensor): Current estimated optical flow.

        Returns:
            Tensor: Feature map by indexing from the correlation pyramid.
        """
        B, _, H, W = flow.shape
        xx = torch.arange(0, W, device=flow.device)
        yy = torch.arange(0, H, device=flow.device)
        grid = coords_grid(B, xx, yy) + flow  # shape N, 2, H, W
        grid = grid.permute(0, 2, 3, 1)  # shape N, H, W, 2

        dx = torch.linspace(
            -self.r, self.r, 2 * self.r + 1, device=flow.device)
        dy = torch.linspace(
            -self.r, self.r, 2 * self.r + 1, device=flow.device)
        delta = torch.stack(torch.meshgrid(dy, dx), axis=-1)
        delta_lvl = delta.view(1, 2 * self.r + 1, 2 * self.r + 1, 2)

        out_corr_pyramid = []
        for i, corr in enumerate(corr_pyramid):
            centroid_lvl = grid.reshape(B * H * W, 1, 1, 2) / 2**i
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sample(corr, coords_lvl, self.mode,
                                   self.padding_mode, self.align_corners)
            corr = corr.view(B, H, W, -1)
            out_corr_pyramid.append(corr)

        out = torch.cat(out_corr_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2).contiguous().float()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule


class BasicConvBlock(BaseModule):
    """Basic convolution block for PWC-Net.

    This module consists of several plain convolution layers.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        num_convs (int): Number of convolution layers. Default: 3.
        stride (int): Whether use stride convolution to downsample
            the input feature map. If stride=2, it only uses stride convolution
            in the first convolution layer to downsample the input feature
            map. Options are 1 or 2. Default: 2.
        dilation (int): Whether use dilated convolution to expand the
            receptive field. Set dilation rate of each convolution layer and
            the dilation rate of the first convolution layer is always 1.
            Default: 1.
        kernel_size (int): Kernel size of each feature level. Default: 3.
        conv_cfg (dict , optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config dict for activation layer in
            ConvModule. Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 num_convs: int = 3,
                 stride: int = 2,
                 dilation: int = 1,
                 kernel_size: int = 3,
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: Optional[dict] = None) -> None:
        super(BasicConvBlock, self).__init__()

        convs = []
        in_channels = in_channels
        for i in range(num_convs):
            k = kernel_size[i] if isinstance(kernel_size,
                                             (tuple, list)) else kernel_size
            out_ch = out_channels[i] if isinstance(out_channels,
                                                   (tuple,
                                                    list)) else out_channels

            convs.append(
                ConvModule(
                    in_channels=in_channels,
                    out_channels=out_ch,
                    kernel_size=k,
                    stride=stride if i == 0 else 1,
                    dilation=1 if i == 0 else dilation,
                    padding=k // 2 if i == 0 else dilation,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channels = out_ch

        self.layers = nn.Sequential(*convs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function."""

        out = self.layers(x)
        return out


class BasicEncoder(BaseModule):
    """A basic pyramid feature extraction sub-network.

    Args:
        in_channels (int): Number of input channels.
        pyramid_levels (Sequence[str]): The list of feature pyramid that are
            the keys for output dict.
        num_convs (Sequence[int]): Numbers of conv layers for each
            pyramid level. Default: (3, 3, 3, 3, 3, 3).
        out_channels (Sequence[int]): List of numbers of output
            channels of each pyramid level.
            Default: (16, 32, 64, 96, 128, 196).
        strides (Sequence[int]): List of strides of each pyramid level.
            Default: (2, 2, 2, 2, 2, 2).
        dilations (Sequence[int]): List of dilation of each pyramid level.
            Default: (1, 1, 1, 1, 1, 1).
        kernel_size (Sequence, int): Kernel size of each feature
            level. Default: 3.
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict): Config dict for each activation layer in ConvModule.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 pyramid_levels: Sequence[str],
                 num_convs: Sequence[int] = (3, 3, 3, 3, 3, 3),
                 out_channels: Sequence[int] = (16, 32, 64, 96, 128, 196),
                 strides: Sequence[int] = (2, 2, 2, 2, 2, 2),
                 dilations: Sequence[int] = (1, 1, 1, 1, 1, 1),
                 kernel_size: Union[Sequence, int] = 3,
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[dict, list]] = None) -> None:
        super().__init__(init_cfg)

        assert len(out_channels) == len(num_convs) == len(strides) == len(
            dilations) == len(pyramid_levels)
        self.in_channels = in_channels
        self.pyramid_levels = pyramid_levels
        self.out_channels = out_channels
        self.num_convs = num_convs
        self.strides = strides
        self.dilations = dilations

        self.kernel_size = kernel_size
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        convs = []
        for i in range(len(out_channels)):
            if isinstance(self.kernel_size, (list, tuple)) and len(
                    self.kernel_size) == len(out_channels):
                kernel_size_ = self.kernel_size[i]
            elif isinstance(self.kernel_size, int):
                kernel_size_ = self.kernel_size
            else:
                TypeError('kernel_size must be list, tuple or int, '
                          f'but got {type(kernel_size)}')

            convs.append(
                self._make_layer(
                    in_channels,
                    out_channels[i],
                    num_convs[i],
                    strides[i],
                    dilations[i],
                    kernel_size=kernel_size_))
            in_channels = out_channels[i][-1] if isinstance(
                out_channels[i], (tuple, list)) else out_channels[i]

        self.layers = nn.Sequential(*convs)

    def _make_layer(self,
                    in_channels: int,
                    out_channel: int,
                    num_convs: int,
                    stride: int,
                    dilation: int,
                    kernel_size: int = 3) -> torch.nn.Module:
        """Make basic block."""
        return BasicConvBlock(
            in_channels=in_channels,
            out_channels=out_channel,
            num_convs=num_convs,
            stride=stride,
            dilation=dilation,
            kernel_size=kernel_size,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg)

    def forward(self, x: torch.Tensor) -> dict:
        """Forward function for BasicEncoder.

        Args:
            x (Tensor): The input data.

        Returns:
            dict: The feature pyramid extracted from input data.
        """
        outs = dict()
        for i, convs_layer in enumerate(self.layers):
            x = convs_layer(x)
            if 'level' + str(i + 1) in self.pyramid_levels:
                outs['level' + str(i + 1)] = x

        return outs



_base_ = [
    '../_base_/models/liteflownet/liteflownet_pre_M2S2R2.py',
    '../_base_/datasets/flyingchairs_320x448.py',
    '../_base_/schedules/liteflownet_pre_300k.py',
    '../_base_/default_runtime.py'
]

optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='Adam', lr=4e-5, weight_decay=0.0004, betas=(0.9, 0.999)))

custom_hooks = [
    dict(
        type='LiteFlowNetStageLoadHook',
        src_level='level3',
        dst_level='level2')
]

# Weights are initialized from model of previous stage
load_from = 'https://download.openmmlab.com/mmflow/liteflownet/liteflownet_pre_M3S3R3_8x1_flyingchairs_320x448.pth'  # noqa



test_dataset_type = 'Sintel'
test_data_root = 'data/Sintel'

img_norm_cfg = dict(mean=[0., 0., 0.], std=[255., 255., 255.], to_rgb=False)

global_transform = dict(
    translates=(0.05, 0.05),
    zoom=(1.0, 1.5),
    shear=(0.86, 1.16),
    rotate=(-10., 10.))

relative_transform = dict(
    translates=(0.00375, 0.00375),
    zoom=(0.985, 1.015),
    shear=(1.0, 1.0),
    rotate=(-1.0, 1.0))

train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', with_occ=True),
    dict(
        type='ColorJitter',
        brightness=0.5,
        contrast=0.5,
        saturation=0.5,
        hue=0.5),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=global_transform,
        relative_transform=relative_transform,
        check_bound=True),
    dict(type='RandomCrop', crop_size=(384, 768)),
    dict(type='PackFlowInputs')
]

test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputResize', exponent=6),
    dict(type='PackFlowInputs')
]

flyingthings3d_subset_bidirection_train = dict(
    type='FlyingThings3DSubset',
    pipeline=train_pipeline,
    data_root='data/FlyingThings3D_subset',
    test_mode=False,
    scene=None)

test_data_cleanpass = dict(
    type=test_dataset_type,
    data_root=test_data_root,
    pipeline=test_pipeline,
    test_mode=False,
    pass_style='clean')

test_data_finalpass = dict(
    type=test_dataset_type,
    data_root=test_data_root,
    pipeline=test_pipeline,
    test_mode=False,
    pass_style='final')

train_dataloader = dict(
    batch_size=1,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    num_workers=2,
    drop_last=True,
    persistent_workers=True,
    dataset=flyingthings3d_subset_bidirection_train)

val_dataloader = [
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=test_data_cleanpass),
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=test_data_finalpass)
]
test_dataloader = val_dataloader
val_evaluator = [
    dict(type='EndPointError', prefix='clean'),
    dict(type='EndPointError', prefix='final')
]
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os.path as osp

from mmengine.utils import mkdir_or_exist
from utils import get_data_filename


def parse_args():
    parser = argparse.ArgumentParser(
        description='KITTI2012 dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/kitti2012',
        help='Directory for dataset.')
    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/kitti2012/',
        help='Directory to save the annotation files for KITTI2012 dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    subset_dir = 'training'

    args.data_root = osp.join(args.data_root, subset_dir)
    # In KITTI 2012, data in `colored_0` is original data
    img1_dir = osp.join(args.data_root, 'colored_0')
    img2_dir = osp.join(args.data_root, 'colored_0')
    flow_dir = osp.join(args.data_root, 'flow_occ')

    img1_suffix = '_10.png',
    img2_suffix = '_11.png',
    flow_suffix = '_10.png'

    img1_filenames = get_data_filename(img1_dir, img1_suffix)
    img2_filenames = get_data_filename(img2_dir, img2_suffix)
    flow_filenames = get_data_filename(flow_dir, flow_suffix)

    data_list = []
    for i_img1, i_img2, i_flow_fw in zip(img1_filenames, img2_filenames,
                                         flow_filenames):
        data_info = dict(
            img1_path=i_img1, img2_path=i_img2, flow_fw_path=i_flow_fw)
        data_list.append(data_info)

    mkdir_or_exist(args.save_dir)
    annotation_file = osp.join(args.save_dir, 'train.json')
    with open(annotation_file, 'w') as jsonfile:
        json.dump({'data_list': data_list, 'metainfo': {}}, jsonfile)


if __name__ == '__main__':
    main()



_base_ = [
    '../_base_/models/liteflownet/liteflownet.py',
    '../_base_/datasets/kitti2012_kitti2015_320x896.py',
    '../_base_/schedules/liteflownet_ft_500k.py',
    '../_base_/default_runtime.py'
]

model = dict(
    decoder=dict(
        flow_loss=dict(
            _delete_=True,
            type='MultiLevelCharbonnierLoss',
            resize_flow='upsample',
            weights=dict(
                level6=0.32,
                level5=0.08,
                level4=0.02,
                level3=0.01,
                level2=0.005),
            q=0.2,
            eps=0.01,
            reduction='sum'),
        init_cfg=None))

# Train on FlyingChairs and FlyingThings3D_subset and finetune on KITTI
load_from = 'https://download.openmmlab.com/mmflow/liteflownet/liteflownet_8x1_500k_flyingthings3d_subset_384x768.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from .corr_lookup import CorrLookup
from .correlation_block import CorrBlock
from .densenet import BasicDenseBlock, DenseLayer
from .estimators_link import BasicLink, LinkOutput
from .misc import unpack_flow_data_samples
from .occlusion_estimation import occlusion_estimation
from .res_layer import BasicBlock, Bottleneck, ResLayer
from .warp import Warp

__all__ = [
    'ResLayer', 'BasicBlock', 'Bottleneck', 'BasicLink', 'LinkOutput',
    'DenseLayer', 'BasicDenseBlock', 'CorrBlock', 'occlusion_estimation',
    'Warp', 'CorrLookup', 'unpack_flow_data_samples'
]



model = dict(
    type='PWCNet',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    encoder=dict(
        type='PWCNetEncoder',
        in_channels=3,
        net_type='Basic',
        pyramid_levels=[
            'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
        ],
        out_channels=(16, 32, 64, 96, 128, 196),
        strides=(2, 2, 2, 2, 2, 2),
        dilations=(1, 1, 1, 1, 1, 1),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
    decoder=dict(
        type='PWCNetDecoder',
        in_channels=dict(
            level6=81, level5=213, level4=181, level3=149, level2=117),
        flow_div=20.,
        corr_cfg=dict(type='Correlation', max_displacement=4, padding=0),
        warp_cfg=dict(type='Warp', align_corners=True, use_mask=True),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        scaled=False,
        post_processor=dict(type='ContextNet', in_channels=565),
        flow_loss=dict(
            type='MultiLevelEPE',
            p=2,
            reduction='sum',
            weights={
                'level2': 0.005,
                'level3': 0.01,
                'level4': 0.02,
                'level5': 0.08,
                'level6': 0.32
            }),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(),
    init_cfg=dict(
        type='Kaiming',
        nonlinearity='leaky_relu',
        layer=['Conv2d', 'ConvTranspose2d'],
        mode='fan_in',
        bias=0))
randomness = dict(seed=0, diff_rank_seed=True)



_base_ = [
    '../_base_/models/maskflownets.py',
    '../_base_/datasets/flyingthings3d_subset_384x768.py',
    '../_base_/schedules/schedule_s_fine.py', '../_base_/default_runtime.py'
]

optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='Adam', lr=0.0001, weight_decay=0., betas=(0.9, 0.999)))
# Train on FlyingChairs and finetune on FlyingThings3D_subset
load_from = 'https://download.openmmlab.com/mmflow/maskflownet/maskflownets_8x1_slong_flyingchairs_384x448.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Tuple

import torch
from torch import Tensor

from mmflow.structures import FlowDataSample


def unpack_flow_data_samples(
        data_samples: Sequence[FlowDataSample]) -> Tuple[Optional[Tensor]]:
    """Unpack data sample list.

    Args:
        data_samples (Sequence[FlowDataSample]): The list of data samples


    Returns:
        Tuple[Optional[Tensor]]: Tuple of ground truth tensor.
    """
    batch_gt_flow_fw = []
    batch_gt_flow_bw = []
    batch_gt_occ_fw = []
    batch_gt_occ_bw = []
    batch_gt_valid_fw = []
    batch_gt_valid_bw = []

    for data_sample in data_samples:
        if hasattr(data_sample, 'gt_flow_fw'):
            batch_gt_flow_fw.append(data_sample.gt_flow_fw.data)
        if hasattr(data_sample, 'gt_flow_bw'):
            batch_gt_flow_bw.append(data_sample.gt_flow_bw.data)
        if hasattr(data_sample, 'gt_occ_fw'):
            batch_gt_occ_fw.append(data_sample.gt_occ_fw.data)
        if hasattr(data_sample, 'gt_occ_bw'):
            batch_gt_occ_bw.append(data_sample.gt_occ_bw.data)
        if hasattr(data_sample, 'gt_valid_fw'):
            batch_gt_valid_bw.append(data_sample.gt_valid_fw.data)
        if hasattr(data_sample, 'gt_valid_bw'):
            batch_gt_valid_bw.append(data_sample.gt_valid_bw.data)

    batch_gt_flow_fw = torch.stack(
        batch_gt_flow_fw) if len(batch_gt_flow_fw) > 0 else None
    batch_gt_flow_bw = torch.stack(
        batch_gt_flow_bw) if len(batch_gt_flow_bw) > 0 else None
    batch_gt_occ_fw = torch.stack(
        batch_gt_occ_fw) if len(batch_gt_occ_fw) > 0 else None
    batch_gt_occ_bw = torch.stack(
        batch_gt_occ_bw) if len(batch_gt_occ_bw) > 0 else None
    batch_gt_valid_fw = torch.cat(
        batch_gt_valid_fw, dim=0) if len(batch_gt_valid_fw) > 0 else None
    batch_gt_valid_bw = torch.cat(
        batch_gt_valid_bw, dim=0) if len(batch_gt_valid_bw) > 0 else None
    return (batch_gt_flow_fw, batch_gt_flow_bw, batch_gt_occ_fw,
            batch_gt_occ_bw, batch_gt_valid_fw, batch_gt_valid_bw)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Union

from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptMultiConfig, OptSampleList, SampleList, TensorDict
from ..builder import build_decoder, build_encoder
from .base_flow_estimator import FlowEstimator


@MODELS.register_module()
class PWCNet(FlowEstimator):
    """PWC-Net model.

    Args:
        encoder (dict): The config of encoder.
        decoder (dict): The config of decoder.
        init_cfg (list, dict, optional): Config of dict weights initialization.
            Default: None.
    """

    def __init__(self,
                 encoder: dict,
                 decoder: dict,
                 init_cfg: OptMultiConfig = None,
                 **kwargs) -> None:

        super().__init__(init_cfg=init_cfg, **kwargs)
        self.encoder = build_encoder(encoder)
        self.decoder = build_decoder(decoder)

    def extract_feat(self, imgs: Tensor) -> TensorDict:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[Dict[str, Tensor], Dict[str, Tensor]]: The feature pyramid of
                the first input image and the feature pyramid of second input
                image.
        """

        in_channels = self.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]
        return self.encoder(img1), self.encoder(img2)

    def loss(self, inputs: Tensor,
             data_samples: SampleList) -> Union[dict, list]:
        """Calculate losses from a batch of inputs and data samples.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is inputs[N, :3, H, W] and img2 is
                inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            dict: A dictionary of loss components.
        """

        return self.decoder.loss(*self.extract_feat(inputs), data_samples)

    def predict(self,
                inputs: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is inputs[N, :3, H, W] and img2 is
                inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        return self.decoder.predict(*self.extract_feat(inputs), data_samples)

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> TensorDict:
        """Network forward process. Usually includes backbone, neck and head
        forward without any post-processing.

        Args:
            inputs (Tensor): Input images of shape (N, 6, H, W).
                img1 is inputs[N, :3, H, W] and img2 is
                inputs[N, 3:, H, W]. These should usually be mean
                centered and std scaled.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
        Returns:
            Dict[str, :obj:`FlowDataSample`]: The predicted optical flow
            from level6 to level2.
        """
        return self.decoder(*self.extract_feat(inputs))



_base_ = [
    '../_base_/models/maskflownet.py',
    '../_base_/datasets/flyingthings3d_subset_384x768.py',
    '../_base_/schedules/maskflownet_500k.py', '../_base_/default_runtime.py'
]

# Train on FlyingChairs and finetune on FlyingThings3D_subset
load_from = 'https://download.openmmlab.com/mmflow/maskflownet/maskflownet_8x1_800k_flyingchairs_384x448.pth'  # noqa



model = dict(
    type='FlowNetCSS',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    flownetC=dict(
        type='FlowNetC',
        freeze_net=True,
        encoder=dict(
            type='FlowNetEncoder',
            in_channels=3,
            pyramid_levels=['level1', 'level2', 'level3'],
            out_channels=(64, 128, 256),
            kernel_size=(7, 5, 5),
            strides=(2, 2, 2),
            num_convs=(1, 1, 1),
            dilations=(1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        corr_level='level3',
        corr_encoder=dict(
            type='CorrEncoder',
            in_channels=473,
            pyramid_levels=['level3', 'level4', 'level5', 'level6'],
            kernel_size=(3, 3, 3, 3),
            num_convs=(1, 2, 2, 2),
            out_channels=(256, 512, 512, 1024),
            redir_in_channels=256,
            redir_channels=32,
            strides=(1, 2, 2, 2),
            dilations=(1, 1, 1, 1),
            corr_cfg=dict(
                type='Correlation',
                kernel_size=1,
                max_displacement=10,
                stride=1,
                padding=0,
                dilation_patch=2),
            scaled=False,
            conv_cfg=None,
            norm_cfg=None,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        decoder=dict(
            type='FlowNetCDecoder',
            in_channels=dict(
                level6=1024, level5=1026, level4=770, level3=386, level2=194),
            out_channels=dict(level6=512, level5=256, level4=128, level3=64),
            deconv_bias=True,
            pred_bias=True,
            upsample_bias=True,
            norm_cfg=None,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict()),
    flownetS1=dict(
        type='FlowNetS',
        freeze_net=True,
        encoder=dict(
            type='FlowNetEncoder',
            in_channels=12,
            pyramid_levels=[
                'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
            ],
            num_convs=(1, 1, 2, 2, 2, 2),
            out_channels=(64, 128, 256, 512, 512, 1024),
            kernel_size=(7, 5, (5, 3), 3, 3, 3),
            strides=(2, 2, 2, 2, 2, 2),
            dilations=(1, 1, 1, 1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        decoder=dict(
            type='FlowNetSDecoder',
            in_channels=dict(
                level6=1024, level5=1026, level4=770, level3=386, level2=194),
            out_channels=dict(level6=512, level5=256, level4=128, level3=64),
            deconv_bias=True,
            pred_bias=True,
            upsample_bias=False,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict()),
    flownetS2=dict(
        type='FlowNetS',
        encoder=dict(
            type='FlowNetEncoder',
            in_channels=12,
            pyramid_levels=[
                'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
            ],
            num_convs=(1, 1, 2, 2, 2, 2),
            out_channels=(64, 128, 256, 512, 512, 1024),
            kernel_size=(7, 5, (5, 3), 3, 3, 3),
            strides=(2, 2, 2, 2, 2, 2),
            dilations=(1, 1, 1, 1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        decoder=dict(
            type='FlowNetSDecoder',
            in_channels=dict(
                level6=1024, level5=1026, level4=770, level3=386, level2=194),
            out_channels=dict(level6=512, level5=256, level4=128, level3=64),
            deconv_bias=True,
            pred_bias=True,
            upsample_bias=False,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            flow_loss=dict(
                type='MultiLevelEPE',
                p=1,
                q=0.4,
                eps=0.01,
                reduction='sum',
                weights={
                    'level2': 0.005,
                    'level3': 0.01,
                    'level4': 0.02,
                    'level5': 0.08,
                    'level6': 0.32
                }),
        ),
        init_cfg=[
            dict(
                type='Kaiming',
                layer=['Conv2d', 'ConvTranspose2d'],
                a=0.1,
                mode='fan_in',
                nonlinearity='leaky_relu',
                bias=0),
            dict(type='Constant', layer='BatchNorm2d', val=1, bias=0)
        ],
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict()),
    link_cfg=dict(scale_factor=4, mode='bilinear'),
    out_level='level2')
randomness = dict(seed=0, diff_rank_seed=True)



model = dict(
    type='RAFT',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[127.5, 127.5, 127.5],
        std=[127.5, 127.5, 127.5],
        bgr_to_rgb=False),
    num_levels=4,
    radius=4,
    cxt_channels=128,
    h_channels=128,
    encoder=dict(
        type='RAFTEncoder',
        in_channels=3,
        out_channels=256,
        net_type='Basic',
        norm_cfg=dict(type='IN'),
        init_cfg=[
            dict(
                type='Kaiming',
                layer=['Conv2d'],
                mode='fan_out',
                nonlinearity='relu'),
            dict(type='Constant', layer=['InstanceNorm2d'], val=1, bias=0)
        ]),
    cxt_encoder=dict(
        type='RAFTEncoder',
        in_channels=3,
        out_channels=256,
        net_type='Basic',
        norm_cfg=dict(type='SyncBN'),
        init_cfg=[
            dict(
                type='Kaiming',
                layer=['Conv2d'],
                mode='fan_out',
                nonlinearity='relu'),
            dict(type='Constant', layer=['SyncBatchNorm2d'], val=1, bias=0)
        ]),
    decoder=dict(
        type='GMADecoder',
        net_type='Basic',
        num_levels=4,
        radius=4,
        iters=12,
        corr_op_cfg=dict(type='CorrLookup', align_corners=True),
        gru_type='SeqConv',
        heads=1,
        motion_channels=128,
        position_only=False,
        max_pos_size=160,
        flow_loss=dict(type='SequenceLoss'),
        act_cfg=dict(type='ReLU')),
    freeze_bn=False,
    train_cfg=dict(),
    test_cfg=dict(),
)



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.runner import autocast
from torch import Tensor

from mmflow.registry import MODELS


def coords_grid(flow: Tensor) -> Tensor:
    """Generate shifted coordinate grid based based input flow.

    Args:
        flow (Tensor): Estimated optical flow.

    Returns:
        Tensor: The coordinate that shifted by input flow and scale in the
            range [-1, 1].
    """
    B, _, H, W = flow.shape
    xx = torch.arange(0, W, device=flow.device, requires_grad=False)
    yy = torch.arange(0, H, device=flow.device, requires_grad=False)
    coords = torch.meshgrid(yy, xx)
    coords = torch.stack(coords[::-1], dim=0).float()
    grid = coords[None].repeat(B, 1, 1, 1) + flow
    grid[:, 0, ...] = grid[:, 0, ...] * 2. / max(W - 1, 1) - 1.
    grid[:, 1, ...] = grid[:, 1, ...] * 2. / max(H - 1, 1) - 1.
    grid = grid.permute(0, 2, 3, 1)
    return grid


@MODELS.register_module()
class Warp(nn.Module):
    """Warping layer to warp feature using optical flow.

    Args:
        mode (str): interpolation mode to calculate output values. Options are
            'bilinear' and 'nearest'. Defaults to 'bilinear'.
        padding_mode (str): padding mode for outside grid values. Options are
            'zero', 'border' and 'reflection'. Defaults to 'zeros'.
        align_corners (bool): If set to True, the extrema (-1 and 1) are
            considered as referring to the center points of the input’s corner
            pixels. If set to False, they are instead considered as referring
            to the corner points of the input’s corner pixels, making the
            sampling more resolution agnostic. Default to False.
    """

    def __init__(self,
                 mode: str = 'bilinear',
                 padding_mode: str = 'zeros',
                 align_corners: bool = False,
                 use_mask: bool = True) -> None:

        super().__init__()
        self.mode = mode
        self.padding_mode = padding_mode
        self.align_corners = align_corners
        self.use_mask = use_mask

    def forward(self, feat: Tensor, flow: Tensor) -> Tensor:
        """Forward function for warp.

        Args:
            feat (Tensor): Input feature
            flow (Tensor): Input optical flow.

        Returns:
            Tensor: The output feature that was generated by warping input
                feature based input flow.
        """
        grid = coords_grid(flow)
        with autocast(enabled=False):
            out = F.grid_sample(
                feat.float(),
                grid,
                mode=self.mode,
                padding_mode=self.padding_mode,
                align_corners=self.align_corners)

        mask = torch.ones_like(feat, requires_grad=False)
        if self.use_mask:
            with autocast(enabled=False):
                mask = F.grid_sample(
                    mask.float(),
                    grid,
                    mode=self.mode,
                    padding_mode=self.padding_mode,
                    align_corners=self.align_corners)
                mask = (mask > 0.9999).float()
        return out * mask

    def __repr__(self):
        s = self.__class__.__name__
        s += f'(mode={self.mode}, '
        s += f'padding_mode={self.padding_mode}, '
        s += f'align_corners={self.align_corners},'
        s += f'use_mask={self.use_mask})'
        return s



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os.path as osp

import numpy as np
from mmengine.utils import mkdir_or_exist, scandir


def parse_args():
    parser = argparse.ArgumentParser(
        description='FlyingChairs dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/FlyingChairs_release',
        help='Directory for dataset.')
    parser.add_argument(
        '--split-file',
        type=str,
        default='data/FlyingChairs_release/FlyingChairs_train_val.txt',
        help='File name of '
        'train-validation split file for FlyingChairs.')

    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/FlyingChairs_release/',
        help='Directory to save the annotation files for FlyingChairs dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    split = np.loadtxt(args.split_file, dtype=np.int32).tolist()
    # unpack FlyingChairs directly, will see `data` subdirctory.
    img1_dir = osp.join(args.data_root, 'data')
    img2_dir = osp.join(args.data_root, 'data')
    flow_dir = osp.join(args.data_root, 'data')

    # data in FlyingChairs dataset has specific suffix
    img1_suffix = '_img1.ppm'
    img2_suffix = '_img2.ppm'
    flow_suffix = '_flow.flo'

    img1_filenames = [f for f in scandir(img1_dir, suffix=img1_suffix)]
    img2_filenames = [f for f in scandir(img2_dir, suffix=img2_suffix)]
    flow_filenames = [f for f in scandir(flow_dir, suffix=flow_suffix)]
    img1_filenames.sort()
    img2_filenames.sort()
    flow_filenames.sort()

    train_list = []
    test_list = []

    for i, flag in enumerate(split):

        data_info = dict(
            img1_path=osp.join(img1_dir, img1_filenames[i]),
            img2_path=osp.join(img2_dir, img2_filenames[i]),
            flow_fw_path=osp.join(flow_dir, flow_filenames[i]))

        if flag == 1:
            train_list.append(data_info)
        else:
            test_list.append(data_info)
    mkdir_or_exist(args.save_dir)
    with open(osp.join(args.save_dir, 'train.json'), 'w') as jsonfile:
        json.dump({'data_list': train_list, 'metainfo': {}}, jsonfile)

    with open(osp.join(args.save_dir, 'test.json'), 'w') as jsonfile:
        json.dump({'data_list': test_list, 'metainfo': {}}, jsonfile)


if __name__ == '__main__':
    main()



_base_ = [
    '../_base_/models/liteflownet/liteflownet_pre_M4S4R4.py',
    '../_base_/datasets/flyingchairs_320x448.py',
    '../_base_/schedules/liteflownet_pre_240k.py',
    '../_base_/default_runtime.py'
]

custom_hooks = [
    dict(
        type='LiteFlowNetStageLoadHook',
        src_level='level5',
        dst_level='level4')
]

# Weights are initialized from model of previous stage
load_from = 'https://download.openmmlab.com/mmflow/liteflownet/liteflownet_pre_M5S5R5_8x1_flyingchairs_320x448.pth'  # noqa



model = dict(
    type='LiteFlowNet',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    encoder=dict(
        type='NetC',
        in_channels=3,
        pyramid_levels=[
            'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
        ],
        out_channels=(32, 32, 64, 96, 128, 192),
        strides=(1, 2, 2, 2, 2, 2),
        num_convs=(1, 3, 2, 2, 1, 1),
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        init_cfg=None),
    decoder=dict(
        type='NetE',
        in_channels=dict(level4=96, level5=128, level6=192),
        corr_channels=dict(level4=49, level5=49, level6=49),
        sin_channels=dict(level4=194, level5=258, level6=386),
        rin_channels=dict(level4=131, level5=131, level6=195),
        feat_channels=64,
        mfeat_channels=(128, 128, 96, 64, 32),
        sfeat_channels=(128, 128, 96, 64, 32),
        rfeat_channels=(128, 128, 64, 64, 32, 32),
        patch_size=dict(level4=5, level5=3, level6=3),
        corr_cfg=dict(
            level4=dict(type='Correlation', max_displacement=3),
            level5=dict(type='Correlation', max_displacement=3),
            level6=dict(type='Correlation', max_displacement=3)),
        warp_cfg=dict(type='Warp', align_corners=True, use_mask=True),
        flow_div=20.,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        scaled_corr=False,
        regularized_flow=True,
        extra_training_loss=False,
        flow_loss=dict(
            type='MultiLevelEPE',
            weights=dict(level6=0.32, level5=0.08, level4=0.02),
            p=2,
            reduction='sum'),
        init_cfg=None),
    # model training and testing settings
    init_cfg=dict(
        type='Kaiming',
        nonlinearity='leaky_relu',
        layer=['Conv2d', 'ConvTranspose2d'],
        mode='fan_in',
        bias=0),
    train_cfg=dict(),
    test_cfg=dict(),
)
randomness = dict(seed=0, diff_rank_seed=True)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os
import os.path as osp
import re
from typing import Sequence

from mmengine.utils import mkdir_or_exist
from utils import get_data_filename

# these files contain nan, so exclude them.
exclude_files = dict(
    left_into_future=[
        '0004573.flo',
        '0006336.flo',
        '0016948.flo',
        '0015148.flo',
        '0006922.flo',
        '0003147.flo',
        '0003149.flo',
        '0000879.flo',
        '0006337.flo',
        '0014658.flo',
        '0015748.flo',
        '0001717.flo',
        '0000119.flo',
        '0017578.flo',
        '0004118.flo',
        '0004117.flo',
        '0004304.flo',
        '0004154.flo',
        '0011530.flo',
    ],
    right_into_future=[
        '0006336.flo',
        '0003148.flo',
        '0004117.flo',
        '0003666.flo',
    ],
    left_into_past=[
        '0000162.flo',
        '0004705.flo',
        '0006878.flo',
        '0004876.flo',
        '0004045.flo',
        '0000053.flo',
        '0005055.flo',
        '0000163.flo',
        '0000161.flo',
        '0000121.flo',
        '0000931.flo',
        '0005054.flo',
    ],
    right_into_past=[
        '0006878.flo',
        '0003147.flo',
        '0001549.flo',
        '0000053.flo',
        '0005034.flo',
        '0003148.flo',
        '0005055.flo',
        '0000161.flo',
        '0001648.flo',
        '0000160.flo',
        '0005054.flo',
    ])


def parse_args():
    parser = argparse.ArgumentParser(
        description='FlyingThings3D_subset dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/FlyingThings3D_subset',
        help='Directory for dataset.')
    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/FlyingThings3D_subset/',
        help='Directory to save the annotation files for '
        'FlyingThings3D subset dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    flow_fw_dir = 'into_future'
    flow_bw_dir = 'into_past'
    flow_suffix = '.flo'
    occ_suffix = '.png'
    img_suffix = '.png'

    def _revise_dir(flow_fw_filename) -> Sequence[str]:
        """Revise directory of optical flow to get the directories of image and
        occlusion mask.

        Args:
            flow_fw_filename (str):  abstract file path of optical flow from
                image1 to image2.
        Returns:
            Sequence[str]: abstract paths of image1, image2, forward
                occlusion mask, backward occlusion mask and backward optical
                flow.
        """

        idx_f = int(osp.splitext(osp.basename(flow_fw_filename))[0])

        img1_filename = flow_fw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}image_clean{os.sep}')
        img1_filename = img1_filename.replace(f'{os.sep}into_future{os.sep}',
                                              f'{os.sep}')

        img1_filename = img1_filename.replace(flow_suffix, img_suffix)
        img2_filename = re.sub(r'\d{7}', f'{idx_f+1:07d}', img1_filename)

        flow_bw_filename = flow_fw_filename.replace(
            f'{os.sep}into_future{os.sep}', f'{os.sep}into_past{os.sep}')
        flow_bw_filename = re.sub(r'\d{7}', f'{idx_f+1:07d}', flow_bw_filename)

        occ_fw_filename = flow_fw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}flow_occlusions{os.sep}')

        occ_fw_filename = occ_fw_filename.replace(flow_suffix, occ_suffix)
        occ_bw_filename = flow_bw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}flow_occlusions{os.sep}')

        occ_bw_filename = occ_bw_filename.replace(flow_suffix, occ_suffix)

        return (img1_filename, img2_filename, occ_fw_filename, occ_bw_filename,
                flow_bw_filename)

    train_list = []
    test_list = []
    for subset in ('train', 'val'):
        data_root = osp.join(args.data_root, subset)

        flow_root = osp.join(data_root, 'flow')

        all_scene = ['left', 'right']
        for scene in all_scene:
            flow_dir = osp.join(flow_root, scene)

            flow_fw_dir = osp.join(flow_dir, 'into_future')
            flow_bw_dir = osp.join(flow_dir, 'into_past')

            tmp_flow_fw_filenames = get_data_filename(
                flow_fw_dir, None, exclude_files[scene + '_into_future'])
            tmp_flow_bw_filenames = get_data_filename(
                flow_bw_dir, None, exclude_files[scene + '_into_past'])

            for ii in range(len(tmp_flow_fw_filenames)):
                flo_fw = tmp_flow_fw_filenames[ii]
                img1, img2, occ_fw, occ_bw, flo_bw = _revise_dir(flo_fw)
                if (not (flo_bw in tmp_flow_bw_filenames)
                        or not osp.isfile(img1) or not osp.isfile(img2)
                        or not osp.isfile(occ_fw) or not osp.isfile(occ_bw)):
                    continue
                data_info = dict(
                    scene=scene,
                    img1_path=img1,
                    img2_path=img2,
                    flow_fw_path=flo_fw,
                    flow_bw_path=flo_bw,
                    occ_fw_path=occ_fw,
                    occ_bw_path=occ_bw)
                if subset == 'train':
                    train_list.append(data_info)
                else:
                    test_list.append(data_info)
    mkdir_or_exist(args.save_dir)
    with open(osp.join(args.save_dir, 'train.json'), 'w') as jsonfile:
        json.dump({'data_list': train_list, 'metainfo': {}}, jsonfile)
    with open(osp.join(args.save_dir, 'test.json'), 'w') as jsonfile:
        json.dump({'data_list': test_list, 'metainfo': {}}, jsonfile)


if __name__ == '__main__':
    main()



FlowNet2css_checkpoint = 'https://download.openmmlab.com/mmflow/flownet2/flownet2css-sd_8x1_sfine_flyingthings3d_subset_chairssdhom_384x448.pth'  # noqa
FlowNet2sd_checkpoint = 'https://download.openmmlab.com/mmflow/flownet2/flownet2sd_8x1_slong_chairssdhom_384x448.pth'  # noqa: E501

model = dict(
    type='FlowNet2',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    flownetCSS=dict(
        type='FlowNetCSS',
        freeze_net=True,
        flownetC=dict(
            type='FlowNetC',
            encoder=dict(
                type='FlowNetEncoder',
                in_channels=3,
                pyramid_levels=['level1', 'level2', 'level3'],
                out_channels=(64, 128, 256),
                kernel_size=(7, 5, 5),
                strides=(2, 2, 2),
                num_convs=(1, 1, 1),
                dilations=(1, 1, 1),
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            ),
            corr_level='level3',
            corr_encoder=dict(
                type='CorrEncoder',
                in_channels=473,
                pyramid_levels=['level3', 'level4', 'level5', 'level6'],
                kernel_size=(3, 3, 3, 3),
                num_convs=(1, 2, 2, 2),
                out_channels=(256, 512, 512, 1024),
                redir_in_channels=256,
                redir_channels=32,
                strides=(1, 2, 2, 2),
                dilations=(1, 1, 1, 1),
                corr_cfg=dict(
                    type='Correlation',
                    kernel_size=1,
                    max_displacement=10,
                    stride=1,
                    padding=0,
                    dilation_patch=2),
                scaled=False,
                conv_cfg=None,
                norm_cfg=None,
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            ),
            decoder=dict(
                type='FlowNetCDecoder',
                in_channels=dict(
                    level6=1024,
                    level5=1026,
                    level4=770,
                    level3=386,
                    level2=194),
                out_channels=dict(
                    level6=512, level5=256, level4=128, level3=64),
                deconv_bias=True,
                pred_bias=True,
                upsample_bias=True,
                norm_cfg=None,
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
            # model training and testing settings
            train_cfg=dict(),
            test_cfg=dict()),
        flownetS1=dict(
            type='FlowNetS',
            encoder=dict(
                type='FlowNetEncoder',
                in_channels=12,
                pyramid_levels=[
                    'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
                ],
                num_convs=(1, 1, 2, 2, 2, 2),
                out_channels=(64, 128, 256, 512, 512, 1024),
                kernel_size=(7, 5, (5, 3), 3, 3, 3),
                strides=(2, 2, 2, 2, 2, 2),
                dilations=(1, 1, 1, 1, 1, 1),
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            ),
            decoder=dict(
                type='FlowNetSDecoder',
                in_channels=dict(
                    level6=1024,
                    level5=1026,
                    level4=770,
                    level3=386,
                    level2=194),
                out_channels=dict(
                    level6=512, level5=256, level4=128, level3=64),
                deconv_bias=True,
                pred_bias=True,
                upsample_bias=False,
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
            # model training and testing settings
            train_cfg=dict(),
            test_cfg=dict()),
        flownetS2=dict(
            type='FlowNetS',
            encoder=dict(
                type='FlowNetEncoder',
                in_channels=12,
                pyramid_levels=[
                    'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
                ],
                num_convs=(1, 1, 2, 2, 2, 2),
                out_channels=(64, 128, 256, 512, 512, 1024),
                kernel_size=(7, 5, (5, 3), 3, 3, 3),
                strides=(2, 2, 2, 2, 2, 2),
                dilations=(1, 1, 1, 1, 1, 1),
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            ),
            decoder=dict(
                type='FlowNetSDecoder',
                in_channels=dict(
                    level6=1024,
                    level5=1026,
                    level4=770,
                    level3=386,
                    level2=194),
                out_channels=dict(
                    level6=512, level5=256, level4=128, level3=64),
                deconv_bias=True,
                pred_bias=True,
                upsample_bias=False,
                act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
            # model training and testing settings
            train_cfg=dict(),
            test_cfg=dict(),
        ),
        link_cfg=dict(scale_factor=4, mode='bilinear'),
        out_level='level2',
        flow_div=20.,
        init_cfg=dict(type='Pretrained', checkpoint=FlowNet2css_checkpoint),
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict(),
    ),
    flownetSD=dict(
        type='FlowNetS',
        freeze_net=True,
        encoder=dict(
            type='FlowNetSDEncoder',
            in_channels=6,
            plugin_channels=64,
            pyramid_levels=[
                'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
            ],
            num_convs=(2, 2, 2, 2, 2, 2),
            out_channels=((64, 128), 128, 256, 512, 512, 1024),
            kernel_size=3,
            strides=(2, 2, 2, 2, 2, 2),
            dilations=(1, 1, 1, 1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        decoder=dict(
            type='FlowNetSDecoder',
            in_channels=dict(
                level6=1024, level5=1026, level4=770, level3=386, level2=194),
            out_channels=dict(level6=512, level5=256, level4=128, level3=64),
            inter_channels=dict(level5=512, level4=256, level3=128, level2=64),
            deconv_bias=True,
            pred_bias=True,
            upsample_bias=True,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        init_cfg=dict(type='Pretrained', checkpoint=FlowNet2sd_checkpoint),
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict(),
    ),
    flownet_fusion=dict(
        type='FlowNetS',
        encoder=dict(
            type='FlowNetEncoder',
            in_channels=11,
            pyramid_levels=['level1', 'level2', 'level3'],
            num_convs=(1, 2, 2),
            out_channels=(64, (64, 128), 128),
            kernel_size=3,
            strides=(1, 2, 2),
            dilations=(1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        decoder=dict(
            type='FlowNetSDecoder',
            in_channels=dict(level3=128, level2=162, level1=82),
            out_channels=dict(level3=32, level2=16),
            inter_channels=dict(level2=32, level1=16),
            deconv_bias=True,
            pred_bias=True,
            upsample_bias=True,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            flow_loss=dict(
                type='MultiLevelEPE',
                p=2,
                reduction='sum',
                flow_div=1.,
                weights=dict(level1=0.005)),
            flow_div=1.,
            init_cfg=[
                dict(
                    type='Xavier',
                    distribution='uniform',
                    layer=['Conv2d', 'ConvTranspose2d'],
                    bias=0),
                dict(type='Constant', layer='BatchNorm2d', val=1, bias=0)
            ])),
    link_cfg=dict(scale_factor=4, mode='nearest'),
    flow_div=20.,
    out_level='level2',
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict())
randomness = dict(seed=0, diff_rank_seed=True)



_base_ = [
    '../_base_/models/raft.py',
    '../_base_/datasets/sintel_cleanx100_sintel_fianlx100_flyingthings3d_raft_368x768.py',  # noqa
    '../_base_/schedules/raft_100k.py',
    '../_base_/default_runtime.py'
]

model = dict(
    decoder=dict(
        type='RAFTDecoder',
        net_type='Basic',
        num_levels=4,
        radius=4,
        iters=12,
        corr_op_cfg=dict(type='CorrLookup', align_corners=True),
        gru_type='SeqConv',
        flow_loss=dict(type='SequenceLoss', gamma=0.85),
        act_cfg=dict(type='ReLU')),
    freeze_bn=True,
    test_cfg=dict(iters=32))

optim_wrapper = dict(optimizer=dict(lr=0.000125, weight_decay=0.00001))
param_scheduler = dict(eta_max=0.000125)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')

# Train on FlyingChairs and FlyingThings3D, and finetune on FlyingThings3D
# and Sintel
load_from = 'https://download.openmmlab.com/mmflow/raft/raft_8x2_100k_flyingthings3d_400x720.pth'  # noqa



_base_ = [
    '../_base_/models/liteflownet/liteflownet.py',
    '../_base_/datasets/flyingthings3d_subset_384x768.py',
    '../_base_/schedules/liteflownet_ft_500k.py',
    '../_base_/default_runtime.py'
]

optimizer = dict(type='Adam', lr=3e-6, weight_decay=0.0004, betas=(0.9, 0.999))
optim_wrapper = dict(type='OptimWrapper', optimizer=optimizer, clip_grad=None)

# Train on FlyingChairs and finetune on FlyingThings3D_subset
load_from = 'https://download.openmmlab.com/mmflow/liteflownet/liteflownet_pre_M2S2R2_8x1_flyingchairs_320x448.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from .multistage import (MultiStageLR, MultiStageMomentum,
                         MultiStageParamScheduler)

__all__ = ['MultiStageLR', 'MultiStageMomentum', 'MultiStageParamScheduler']



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn.functional as F
from torch import Tensor


def census_loss(image1: Tensor, image2: Tensor, patch_size: int = 3) -> Tensor:
    """The census loss.

    Modified from
    https://github.com/lliuz/ARFlow/blob/master/losses/flow_loss.py
    licensed under MIT License,

    Args:
        image1 (Tensor): Tensor of shape (B, 3, H, W)
        image2 (Tensor): Tensor of shape (B, 3, H, W)
        patch_size (int): The window size for census transform. Defaults to 3.

    Returns:
        Tensor: The scaler of census loss between image1 and image2.
    """

    def _rgb_to_grayscale(image):
        grayscale = image[:, 0, :, :] * 0.2989 + \
                    image[:, 1, :, :] * 0.5870 + \
                    image[:, 2, :, :] * 0.1140
        return grayscale[:, None, ...]

    def _census_transform(image):
        intensities = _rgb_to_grayscale(image) * 255
        out_channels = patch_size * patch_size
        weight = torch.eye(out_channels).view(
            (out_channels, 1, patch_size, patch_size)).to(image)
        patches = F.conv2d(intensities, weight, padding=patch_size // 2)
        transf = patches - intensities
        transf_norm = transf / torch.sqrt(0.81 + torch.pow(transf, 2))
        return transf_norm

    def _hamming_distance(t1, t2):
        dist = (t1 - t2)**2
        dist_norm = dist / (0.1 + dist)
        dist_mean = torch.mean(dist_norm, 1, keepdim=True)  # instead of sum
        return dist_mean

    def _valid_mask(t, padding):
        n, _, h, w = t.size()
        inner = torch.ones(n, 1, h - 2 * padding, w - 2 * padding).to(t)
        # just mask the edge
        mask = F.pad(inner, [padding] * 4)
        return mask

    t1 = _census_transform(image1)
    t2 = _census_transform(image2)
    dist = _hamming_distance(t1, t2)
    mask = _valid_mask(image1, patch_size // 2)

    return (dist * mask).sum() / mask.sum()



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os
import os.path as osp

from mmengine.utils import mkdir_or_exist
from utils import get_data_filename


def parse_args():
    parser = argparse.ArgumentParser(description='Sintel dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/Sintel',
        help='Directory for dataset.')
    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/Sintel/',
        help='Directory to save the annotation files for Sintel dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    all_pass = ['clean', 'final']

    def _get_filenames(data_dir, data_suffix, img_idx=None):
        data_filenames = []

        data_filenames = get_data_filename(data_dir, data_suffix)
        data_filenames.sort()
        if img_idx == 1:
            return data_filenames[:-1]
        elif img_idx == 2:
            return data_filenames[1:]
        else:
            return data_filenames

    def _get_data_list(subset_dir):
        """Get the paths for images and optical flow."""
        img_suffix = '.png'
        flow_suffix = '.flo'
        occ_suffix = '.png'
        invalid_suffix = '.png'

        data_root = osp.join(args.data_root, subset_dir)
        if subset_dir == 'testing':
            data_root = osp.join(data_root, 'test')
        flow_root = osp.join(data_root, 'flow')
        occ_root = osp.join(data_root, 'occlusions')
        invalid_root = osp.join(data_root, 'invalid')

        data_list = []

        for p in all_pass:
            img_root = osp.join(data_root, p)
            scene = os.listdir(img_root)
            for s in scene:
                img1_dir = osp.join(img_root, s)
                img2_dir = osp.join(img_root, s)

                flow_dir = osp.join(flow_root, s)
                occ_dir = osp.join(occ_root, s)
                invalid_dir = osp.join(invalid_root, s)
                img1_filenames = _get_filenames(img1_dir, img_suffix, 1)
                img2_filenames = _get_filenames(img2_dir, img_suffix, 2)
                if subset_dir == 'training':
                    flow_filenames = _get_filenames(flow_dir, flow_suffix)
                    occ_filenames = _get_filenames(occ_dir, occ_suffix)
                    invalid_filenames = _get_filenames(invalid_dir,
                                                       invalid_suffix)
                else:
                    flow_filenames = [None] * len(img1_filenames)
                    occ_filenames = [None] * len(img1_filenames)
                    invalid_filenames = [None] * len(img1_filenames)

                for i_img1, i_img2, i_flow, i_occ, i_invalid in zip(
                        img1_filenames, img2_filenames, flow_filenames,
                        occ_filenames, invalid_filenames):
                    data_info = dict(
                        pass_style=p,
                        scene=s,
                        img1_path=i_img1,
                        img2_path=i_img2,
                        flow_fw_path=i_flow,
                        invalid_path=i_invalid,
                        occ_fw_path=i_occ)
                    data_list.append(data_info)
        mkdir_or_exist(args.save_dir)
        if subset_dir == 'training':
            annotation_file = osp.join(args.save_dir, 'train.json')
        else:
            annotation_file = osp.join(args.save_dir, 'test.json')
        with open(annotation_file, 'w') as jsonfile:
            json.dump({'data_list': data_list, 'metainfo': {}}, jsonfile)

    _get_data_list('training')
    _get_data_list('testing')


if __name__ == '__main__':
    main()



_base_ = [
    '../_base_/models/gma/gma_plus-p.py',
    '../_base_/datasets/flyingthings3d_raft_400x720.py',
    '../_base_/schedules/gma_120k.py', '../_base_/default_runtime.py'
]

model = dict(freeze_bn=False, test_cfg=dict(iters=32))

lr = 0.000125
optim_wrapper = dict(optimizer=dict(lr=lr))
param_scheduler = dict(eta_max=lr)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')

# Train on FlyingChairs and finetune on FlyingThings3D
load_from = 'https://download.openmmlab.com/mmflow/gma/gma_plus-p_8x2_120k_flyingchairs_368x496.pth'  # noqa



_base_ = [
    '../_base_/models/irrpwc.py',
    '../_base_/datasets/flyingchairsocc_bi_occ_384x448.py',
    '../_base_/schedules/schedule_s_short.py', '../_base_/default_runtime.py'
]

custom_hooks = [dict(type='EMAHook')]
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')



# Copyright (c) OpenMMLab. All rights reserved.
from .liteflownet_stage_loading import LiteFlowNetStageLoadHook
from .visualization_hook import FlowVisualizationHook

__all__ = ['LiteFlowNetStageLoadHook', 'FlowVisualizationHook']



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

import torch
import torch.nn as nn
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from .raft_decoder import ConvGRU, RAFTDecoder


class RelPosEmb(BaseModule):
    """Relative position embedding.

    Separate embedding vectors are learned for the vertical and horizontal
    offset and are added on query tensor.

    Args:
        max_pos_size (int): maximum position length.
        head_channels (int): The channels of head feature.
    """

    def __init__(self, max_pos_size: int, head_channels: int):
        super().__init__()
        self.rel_height = nn.Embedding(2 * max_pos_size - 1, head_channels)
        self.rel_width = nn.Embedding(2 * max_pos_size - 1, head_channels)
        deltas = torch.arange(max_pos_size).view(
            1, -1) - torch.arange(max_pos_size).view(-1, 1)
        rel_ind = deltas + max_pos_size - 1
        self.register_buffer('rel_ind', rel_ind)

    def forward(self, q: torch.Tensor):
        """forward function for relative position embedding.

        Args:
            q (torch.Tensor): query tensor in attention.

        Returns:
            torch.Tensor: Position score with shape (B, heads, H * W, H * W).
        """
        B, heads, H, W, head_channels = q.shape

        height_emb = self.rel_height(self.rel_ind[:H, :H].reshape(-1)).reshape(
            H, H, -1, head_channels)
        width_emb = self.rel_width(self.rel_ind[:W, :W].reshape(-1)).reshape(
            W, -1, W, head_channels)

        height_score = torch.einsum('b h x y d, x u v d -> b h x y u v', q,
                                    height_emb)
        width_score = torch.einsum('b h x y d, y u v d -> b h x y u v', q,
                                   width_emb)

        return (height_score + width_score).reshape(B, heads, H * W, H * W)


class Attention(BaseModule):
    """Compute 4D attention matrix encodes self-similarity in appearance
    feature space by using context features.

    Args:
        in_channels (int): The channels of input context features
        heads (int): The number of parallel attention heads.
        head_channels (int): The channels of head feature.
        position_only (bool): Whether use position-only attention. Default to
            False.
        max_pos_size (int, optional): The max size of positional embedding
            vectors. If max_pos_size is None, attention is content-only
            self-similarity attention model. If position_only=True,
            max_pos_size must be defined. Default to None.

        Note:
            Attention module has 3 mode based on different `position_only` and
            `max_pos_size`:
            1. If `position_only` is True and `max_pos_size` is defined,
            it denotes the position-only attention module.
            2. If `position_only` is False and `max_pos_size` is defined,
            it denotes the joint position and content-wise attention module.
            3. If `position_only` is False and `max_pos_size` is not defined,
            it denotes the content-only self-similarity attention module.
    """

    def __init__(self,
                 in_channels: int,
                 heads: int,
                 head_channels: int,
                 position_only: bool = False,
                 max_pos_size: Optional[int] = None) -> None:
        super().__init__()
        if position_only and max_pos_size is None:
            raise RuntimeError(
                'Must define `max_pos_size`, if position_only=True')

        self.in_channels = in_channels
        self.heads = heads
        self.head_channels = head_channels
        self.position_only = position_only
        self.max_pos_size = max_pos_size

        self.scale = head_channels**-0.5
        self.to_qk = nn.Conv2d(
            in_channels=in_channels,
            out_channels=self.heads * self.head_channels * 2,
            kernel_size=1,
            bias=False)

        if self.max_pos_size is not None:
            self.pos_emb = RelPosEmb(
                max_pos_size=self.max_pos_size, head_channels=head_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function for computing self-similarity with context
        features.

        Args:
            x (torch.Tensor): The context features.

        Returns:
            torch.Tensor: self-similarity for context features.
        """
        B, _, H, W = x.shape
        q, k = torch.split(
            self.to_qk(x),
            [self.heads * self.head_channels, self.heads * self.head_channels],
            dim=1)

        # q shape is (B, heads, HxW, head_channels)
        q = q.view(B, self.heads, self.head_channels, H,
                   W).permute(0, 1, 3, 4, 2)
        # k shape is (B, heads, head_channels, HxW)
        k = k.view(B, self.heads, self.head_channels, -1)

        self_similarity = torch.zeros((B, self.heads, H * W, H * W)).to(q)

        if self.max_pos_size is not None:
            self_similarity += self.pos_emb(q)

        if not self.position_only:
            # self_similarity shape is (B, heads, HxW, HxW)
            self_similarity += torch.matmul(
                q.view(B, self.heads, -1, self.head_channels),
                k.view(B, self.heads, self.head_channels, -1)) * self.scale

        attn = self_similarity.softmax(dim=-1)

        return attn


class Aggregate(BaseModule):
    """Computing aggregated global motion features.

    Args:
        in_channels (int): The channels of motion features.
        heads (int): The number of parallel heads.
        head_channels (int): The channels of head feature.
    """

    def __init__(self, in_channels: int, heads: int,
                 head_channels: int) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.heads = heads
        self.head_channels = head_channels

        self.scale = head_channels**-0.5
        self.to_v = nn.Conv2d(
            in_channels=in_channels,
            out_channels=self.heads * self.head_channels,
            kernel_size=1,
            bias=False)
        self.gamma = nn.Parameter(torch.zeros(1))

        if in_channels != heads * head_channels:
            self.project = nn.Conv2d(
                in_channels=heads * head_channels,
                out_channels=in_channels,
                kernel_size=1,
                bias=False)
        else:
            self.project = nn.Sequential()

    def forward(self, attn: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """Forward function to compute aggregated global motion features.

        Args:
            attn (torch.Tensor): The attention features
            x (torch.Tensor): The motion features

        Returns:
            torch.Tensor: The aggregated global motion features
        """
        B, _, H, W = x.shape

        # v shape : B, Heads, Head_channels, HxW
        v = self.to_v(x).view(B, self.heads, self.head_channels, -1)

        # attn shape: B, Heads, HxW, HxW
        # out shape: B, Heads, HxW, Head_channels
        out = torch.matmul(attn, v.permute(0, 1, 3, 2))

        # out shape: B, in_channels, H, W
        out = self.project(out.permute(0, 1, 3, 2).reshape(B, -1, H, W))

        out = x + self.gamma * out

        return out


@MODELS.register_module()
class GMADecoder(RAFTDecoder):
    """The decoder of GMA.

    Args:
        heads (int): The number of parallel attention heads.
        motion_channels (int): The channels of motion channels.
        position_only (bool): Whether use position-only attention. Default to
            False.
        max_pos_size (int, optional): The max size of positional embedding
            vectors. If max_pos_size is None, attention is content-only
            self-similarity attention model. If position_only=True,
            max_pos_size must be defined. Default to None.

        Note:
            Attention module has 3 mode based on different `position_only` and
            `max_pos_size`:
            1. If `position_only` is True and `max_pos_size` is defined,
            it denotes the position-only attention module.
            2. If `position_only` is False and `max_pos_size` is defined,
            it denotes the joint position and content-wise attention module.
            3. If `position_only` is False and `max_pos_size` is not defined,
            it denotes the content-only self-similarity attention module.
    """

    def __init__(self,
                 *args,
                 heads: int,
                 motion_channels: int,
                 position_only: bool = False,
                 max_pos_size: Optional[int] = None,
                 **kwargs) -> None:
        self.heads = heads
        self.motion_channels = motion_channels
        self.position_only = position_only
        self.max_pos_size = max_pos_size

        super().__init__(*args, **kwargs)
        self.attn = Attention(
            in_channels=self.cxt_channels,
            heads=heads,
            head_channels=self.cxt_channels,
            position_only=self.position_only,
            max_pos_size=self.max_pos_size)

        self.aggregator = Aggregate(
            in_channels=motion_channels,
            heads=heads,
            head_channels=motion_channels)

    def make_gru_block(self) -> torch.nn.Module:
        return ConvGRU(
            self.h_channels,
            self.cxt_channels + self.motion_channels * 2,
            net_type=self.gru_type)

    def forward(self, feat1: Tensor, feat2: Tensor, flow: Tensor,
                h_feat: Tensor, cxt_feat: Tensor) -> List[Tensor]:
        """Forward function for RAFTDecoder.

        Args:
            feat1 (Tensor): The feature from the first input image.
            feat2 (Tensor): The feature from the second input image.
            flow (Tensor): The initialized flow when warm start.
            h_feat (Tensor): The hidden state for GRU cell.
            cxt_feat (Tensor): The contextual feature from the first image.

        Returns:
            Sequence[Tensor]: The list of predicted optical flow.
        """

        corr_pyramid = self.corr_block(feat1, feat2)
        upflow_preds = []
        delta_flow = torch.zeros_like(flow)

        attention = self.attn(cxt_feat)

        for _ in range(self.iters):
            flow = flow.detach()
            corr = self.corr_lookup(corr_pyramid, flow)
            motion_feat = self.encoder(corr, flow)
            motion_features_global = self.aggregator(attention, motion_feat)
            x = torch.cat([cxt_feat, motion_feat, motion_features_global],
                          dim=1)
            h_feat = self.gru(h_feat, x)

            delta_flow = self.flow_pred(h_feat)
            flow = flow + delta_flow

            if hasattr(self, 'mask_pred'):
                mask = .25 * self.mask_pred(h_feat)
            else:
                mask = None

            upflow = self._upsample(flow, mask)
            upflow_preds.append(upflow)

        return upflow_preds



# Copyright (c) OpenMMLab. All rights reserved.
from .misc import sync_random_seed
from .set_env import register_all_modules
from .typing import (ConfigType, MultiConfig, OptConfigType, OptMultiConfig,
                     OptSampleList, SampleList, TensorDict, TensorList)

__all__ = [
    'register_all_modules', 'ConfigType', 'OptConfigType', 'MultiConfig',
    'OptMultiConfig', 'SampleList', 'OptSampleList', 'TensorDict',
    'TensorList', 'sync_random_seed'
]



crop_size = (384, 768)

global_transform = dict(
    translates=(0.05, 0.05),
    zoom=(1.0, 1.5),
    shear=(0.86, 1.16),
    rotate=(-10., 10.))

relative_transform = dict(
    translates=(0.00375, 0.00375),
    zoom=(0.985, 1.015),
    shear=(1.0, 1.0),
    rotate=(-1.0, 1.0))

sintel_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(
        type='ColorJitter',
        brightness=0.5,
        contrast=0.5,
        saturation=0.5,
        hue=0.5),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=global_transform,
        relative_transform=relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='PackFlowInputs')
]

sintel_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputResize', exponent=4),
    dict(type='PackFlowInputs')
]

sintel_clean_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')

sintel_final_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

sintel_clean_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='clean')

sintel_final_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    pass_style='final')

train_dataloader = dict(
    batch_size=1,
    num_workers=5,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=dict(
        type='ConcatDataset',
        datasets=[sintel_clean_train, sintel_final_train]))
val_dataloader = [
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_clean_test),
    dict(
        batch_size=1,
        num_workers=2,
        sampler=dict(type='DefaultSampler', shuffle=False),
        drop_last=False,
        persistent_workers=True,
        dataset=sintel_final_test)
]
test_dataloader = val_dataloader

val_evaluator = [
    dict(type='EndPointError', prefix='clean'),
    dict(type='EndPointError', prefix='final')
]
test_evaluator = val_evaluator
